#  Copyright 2022 Feedzai
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Callable, List, Union
import numpy as np
import pandas as pd
from timeshap.explainer.kernel import TimeShapKernel
import os
import re
import csv
from pathlib import Path
from timeshap.utils import convert_to_indexes, convert_data_to_3d
from timeshap.explainer import temp_coalition_pruning
from timeshap.utils import get_tolerances_to_test


def event_level(f: Callable,
                data: np.array,
                baseline: Union[np.ndarray, pd.DataFrame],
                pruned_idx: int,
                random_seed: int,
                nsamples: int,
                display_events: List[str] = None,
                ) -> pd.DataFrame:
    """Method to calculate event level explanations

    Parameters
    ----------
    f: Callable[[np.ndarray], np.ndarray]
        Point of entry for model being explained.
        This method receives a 3-D np.ndarray (#samples, #seq_len, #features).
        This method returns a 2-D np.ndarray (#samples, 1).

    data: np.array
        Sequence to explain.

    baseline: Union[np.ndarray, pd.DataFrame],
        Dataset baseline. Median/Mean of numerical features and mode of categorical.
        In case of np.array feature are assumed to be in order with `model_features`.
        The baseline can be an average event or an average sequence

    pruned_idx: int
        Index to prune the sequence. All events up to this index are grouped

    random_seed: int
        Used random seed for the sampling process.

    nsamples: int
        The number of coalitions for TimeSHAP to sample.

    display_events: List[str]
        In-order list of event names to be displayed

    Returns
    -------
    pd.DataFrame
    """
    explainer = TimeShapKernel(f, baseline, random_seed, "event")
    shap_values = explainer.shap_values(data, pruning_idx=pruned_idx, nsamples=nsamples)

    if display_events is None:
        display_events = ["Event {}".format(str(-int(i))) for i in np.arange(1, data.shape[1]-pruned_idx+1)]
    else:
        display_events = display_events[-len(shap_values)+1:]
    if pruned_idx > 0:
        display_events += ["Pruned Events"]

    ret_data = []
    for exp, event in zip(shap_values, display_events):
        ret_data += [[random_seed, nsamples, event, exp]]
    return pd.DataFrame(ret_data, columns=['Random seed', 'NSamples', 'Feature', 'Shapley Value'])


def local_event(f: Callable[[np.ndarray], np.ndarray],
                data: np.array,
                event_dict: dict,
                entity_uuid: Union[str, int, float],
                entity_col: str,
                baseline: Union[pd.DataFrame, np.array],
                pruned_idx: int,
                ) -> pd.DataFrame:
    """Method to calculate event level explanations or load them if path is provided

    Parameters
    ----------
    f: Callable[[np.ndarray], np.ndarray]
        Point of entry for model being explained.
        This method receives a 3-D np.ndarray (#samples, #seq_len, #features).
        This method returns a 2-D np.ndarray (#samples, 1).

    data: np.array
        Sequence to explain.

    event_dict: dict
        Information required for the event level explanation calculation

    entity_uuid: Union[str, int, float]
        The indentifier of the sequence that is being pruned.
        Used when fetching information from a csv of explanations

    entity_col: str
        Entity column to identify sequences

    baseline: Union[np.ndarray, pd.DataFrame],
        Dataset baseline. Median/Mean of numerical features and mode of categorical.
        In case of np.array feature are assumed to be in order with `model_features`.
        The baseline can be an average event or an average sequence

    pruned_idx: int
        Index to prune the sequence. All events up to this index are grouped

    Returns
    -------
    pd.DataFrame
    """
    if event_dict.get("path") is None or not os.path.exists(event_dict.get("path")):
        #print("No path to event data provided. Calculating data")
        event_data = event_level(f, data, baseline, pruned_idx, event_dict.get("rs"), event_dict.get("nsamples"))
        if event_dict.get("path") is not None:
            # create directory
            if '/' in event_dict.get("path"):
                Path(event_dict.get("path").rsplit("/", 1)[0]).mkdir(parents=True, exist_ok=True)
            event_data.to_csv(event_dict.get("path"), index=False)
    elif event_dict.get("path") is not None and os.path.exists(event_dict.get("path")):
        event_data = pd.read_csv(event_dict.get("path"))
        if len(event_data.columns) == 5 and entity_col is not None:
            event_data = event_data[event_data[entity_col] == entity_uuid]
        elif len(event_data.columns) == 4:
            pass
        else:
            # TODO
            # the provided csv should be generated by timeshap, by either
            # explaining the whole dataset with TODO or just the instance in question
            raise ValueError
    else:
        raise ValueError

    return event_data


def verify_event_dict(event_dict: dict):
    """Verifies the format of the event dict for event level explanations

    Parameters
    ----------
    event_dict: dict
    """
    if event_dict.get('path'):
        assert isinstance(event_dict.get('path'), str), "Provided path must be a string"

    if event_dict.get('rs', False):
        if isinstance(event_dict.get('rs'), int):
            event_dict['rs'] = [event_dict.get('rs')]
        elif isinstance(event_dict.get('rs'), list):
            assert np.array([isinstance(x, int) for x in event_dict.get('rs')]).all(), "All provided random seeds must be ints."
        else:
            raise ValueError("Unsuported format of random seeds(s). Please provide one seed or a list of them.")
    else:
        print("No random seed provided for event-level explanations. Using default: 42")
        event_dict['rs'] = [42]

    if event_dict.get('nsamples', False):
        if isinstance(event_dict.get('nsamples'), int):
            event_dict['nsamples'] = [event_dict.get('nsamples')]
        elif isinstance(event_dict.get('nsamples'), list):
            assert np.array([isinstance(x, int) for x in event_dict.get('nsamples')]).all(), "All provided nsamples must be ints."
        else:
            raise ValueError("Unsuported format of nsamples. Please provide value or a list of them.")
    else:
        print("No nsamples provided for event-level explanations. Using default: 32000")
        event_dict['nsamples'] = [32000]

    if event_dict.get('tol', False):
        tolerances = event_dict.get('tol')
        if isinstance(tolerances, float):
            event_dict['tol'] = [tolerances]
        elif isinstance(tolerances, list):
            assert np.array([isinstance(x, float) for x in tolerances]).all(), "All provided tolerances must be floats."


def event_explain_all(f: Callable,
                      data: Union[List[np.ndarray], pd.DataFrame, np.array],
                      event_dict: dict,
                      pruning_data: pd.DataFrame = None,
                      baseline: Union[pd.DataFrame, np.array] = None,
                      model_features: List[Union[int, str]] = None,
                      schema: List[str] = None,
                      entity_col: Union[int, str] = None,
                      time_col: Union[int, str] = None,
                      append_to_files: bool = False,
                      verbose: bool = False,
                      ) -> pd.DataFrame:
    """Calculates event level explanations for all entities on the provided
    DataFrame applying pruning if explicit

    Parameters
    ----------
    f: Callable[[np.ndarray], np.ndarray]
        Point of entry for model being explained.
        This method receives a 3-D np.ndarray (#samples, #seq_len, #features).
        This method returns a 2-D np.ndarray (#samples, 1).

    data: Union[List[np.ndarray], pd.DataFrame, np.array]
        Sequences to be explained.
        Must contain columns with names disclosed on `model_features`.

    event_dict: dict
        Information required for the event level explanation calculation

    pruning_data: pd.DataFrame
        Pruning indexes for all sequences being explained.
        Produced by `prune_all`

    baseline: Union[np.ndarray, pd.DataFrame],
        Dataset baseline. Median/Mean of numerical features and mode of categorical.
        In case of np.array feature are assumed to be in order with `model_features`.
        The baseline can be an average event or an average sequence

    model_features: List[str]
        Features to be used by the model. Requires same order as training dataset

    schema: List[str]
        Schema of provided data

    entity_col: str
        Entity column to identify sequences

    time_col: str
        Data column that represents the time feature in order to sort sequences
        temporally

    append_to_files: bool
        Append explanations to files if file already exists

    verbose: bool
        If process is verbose

    Returns
    -------
    pd.DataFrame
    """
    if schema is None and isinstance(data, pd.DataFrame):
        schema = list(data.columns)
    verify_event_dict(event_dict)
    file_path = event_dict.get('path')
    make_predictions = True
    event_data = None

    tolerances_to_calc = get_tolerances_to_test(pruning_data, event_dict)

    if file_path is not None and os.path.exists(file_path) and not append_to_files:
        event_data = pd.read_csv(file_path)
        make_predictions = False

        # TODO resume explanations
        # conditions = []
        # necessary_entities = set(np.unique(data[entity_col].values))
        # event_data = pd.read_csv(file_path)
        # present_entities = set(np.unique(event_data[entity_col].values))
        # if necessary_entities.issubset(present_entities):
        #     conditions.append(True)
        #     event_data = event_data[event_data[entity_col].isin(necessary_entities)]
        #
        # necessary_tols = set(tolerances_to_calc)
        # loaded_csv = pd.read_csv(file_path)
        # present_tols = set(np.unique(loaded_csv['Tolerance'].values))
        # if necessary_tols.issubset(present_tols):
        #     conditions.append(True)
        #     event_data = event_data[event_data['Tolerance'].isin(necessary_tols)]
        #
        # make_predictions = ~np.array(conditions).all()

    if make_predictions:
        random_seeds = list(np.unique(event_dict.get('rs')))
        nsamples = list(np.unique(event_dict.get('nsamples')))
        names = ["Random Seed", "NSamples", "Event", "Shapley Value", "t (event index)", "Entity", 'Tolerance']

        if file_path is not None:
            if os.path.exists(file_path):
                assert append_to_files, "The defined path for event explanations already exists and the append option is turned off. If you wish to append the explanations please use the flag `append_to_files`, otherwise change the provided path."
            else:
                if '/' in file_path:
                    Path(file_path.rsplit("/", 1)[0]).mkdir(parents=True, exist_ok=True)
                with open(file_path, 'w', newline='') as file:
                    writer = csv.writer(file, delimiter=',')
                    writer.writerow(names)

        if time_col is None:
            print("No time col provided, assuming dataset is ordered ascendingly by date")

        model_features_index, entity_col_index, time_col_index = convert_to_indexes(model_features, schema, entity_col, time_col)
        data = convert_data_to_3d(data, entity_col_index, time_col_index)

        ret_event_data = []
        for rs in random_seeds:
            for ns in nsamples:
                for sequence in data:
                    if entity_col is not None:
                        entity = sequence[0, 0, entity_col_index]
                    if model_features:
                        sequence = sequence[:, :, model_features]
                    sequence = sequence.astype(np.float64)
                    event_data = None
                    prev_pruning_idx = None
                    for tol in tolerances_to_calc:
                        if pruning_data is None:
                            #we need to perform the pruning on the fly
                            coal_prun_idx, _ = temp_coalition_pruning(f, sequence, baseline, tol)
                            pruning_idx = data.shape[1] + coal_prun_idx
                        else:
                            instance = pruning_data[pruning_data["Entity"] == entity]
                            pruning_idx = instance[instance['Tolerance'] == tol]['Pruning idx'].iloc[0]
                            pruning_idx = sequence.shape[1] + pruning_idx

                        if prev_pruning_idx == pruning_idx:
                            # we have already calculated this, let's use it from the last iteration
                            event_data['Tolerance'] = tol
                        else:
                            local_event_dict = {'rs': rs, 'nsamples': ns}
                            event_data = local_event(f, sequence, local_event_dict, entity, entity_col, baseline, pruning_idx)
                            event_data['Event index'] = event_data['Feature'].apply(lambda x: 1 if x == 'Pruned Events' else -int(re.findall(r'\d+', x)[0])+1)
                            event_data[entity_col] = entity
                            event_data['Tolerance'] = tol

                        if file_path is not None:
                            with open(file_path, 'a', newline='') as file:
                                writer = csv.writer(file, delimiter=',')
                                writer.writerows(event_data.values)
                        ret_event_data.append(event_data.values)
                        prev_pruning_idx = pruning_idx

        event_data = pd.DataFrame(np.concatenate(ret_event_data), columns=names)
        event_data = event_data.astype({'NSamples': 'int', 'Random Seed': 'int', 'Tolerance': 'float', 'Shapley Value': 'float', 't (event index)': 'int'})
    return event_data



#  Copyright 2022 Feedzai
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Callable, Union, Tuple, List
import numpy as np
import pandas as pd
from timeshap.explainer.kernel import TimeShapKernel
import os
import csv
from pathlib import Path
from timeshap.utils import convert_to_indexes, convert_data_to_3d


def calc_prun_indexes(df: pd.DataFrame,
                      tol: Union[float, int, list]
                      ) -> pd.DataFrame:
    """Calculates the pruning indexes given pruning data and tolerances

    Parameters
    ----------
    df: pd.DataFrame
        Pruning data to be analysed produced by `prune_all`

    tol: Union[float, list]
        The tolerances to analyze the pruning

    Returns
    -------
    pd.DataFrame
    """
    if "Tolerance" not in list(df.columns):
        pruning_data = []
        if not isinstance(tol, list):
            tol = [tol]
        for uuid in np.unique(df.iloc[:, -1].values):
            uuid_data = df[df.iloc[:, -1] == uuid]
            pruning_data.append([uuid, -1, -(uuid_data.shape[0] / 2) + 1])
            for tolerance in tol:
                pruning_idx = prune_given_data(uuid_data, tolerance)
                pruning_data.append([uuid, tolerance, pruning_idx])

        df = pd.DataFrame(pruning_data, columns=["Entity", 'Tolerance', 'Pruning idx'])
    return df


def pruning_statistics(df: pd.DataFrame,
                       tol: Union[float, list],
                       ) -> pd.DataFrame:
    """Calculates global pruning statistics with the given tolerances.

    Parameters
    ----------
    df: pd.DataFrame
        Pruning data to be analysed produced by `prune_all`

    tol: Union[float, list]
        The tolerances to analyze the pruning

    Returns
    -------
    pd.DataFrame
    """
    if isinstance(tol, float):
        tol = [tol]
    if "Tolerance" not in list(df.columns):
        df = calc_prun_indexes(df, tol)

    resume = []
    orig = df[df['Tolerance'] == -1]
    for idx, row in orig.iterrows():
        resume += [["Original", 'No Pruning', row["Entity"], -row['Pruning idx']]]

    for tol in tol:
        tolerance_sequences = df[df['Tolerance'] == tol]
        for idx, row in tolerance_sequences.iterrows():
            resume.append(["Pruning", tol,  row["Entity"], -row['Pruning idx']])

    resume_df = pd.DataFrame(resume, columns=["Algorithm", "Tolerance", "Entity", "Sequence Length"])
    resume_df['Mean'] = resume_df['Sequence Length']
    resume_df['Std'] = resume_df['Sequence Length']
    resume_df = resume_df.groupby("Tolerance").agg({"Mean": "mean", "Std": "std"})
    resume_df.reset_index(inplace=True)
    resume_df = resume_df.rename(columns={'index': 'Tolerance'})
    return resume_df


def prune_given_data(data: pd.DataFrame,
                     tolerance: float,
                     ) -> int:
    """Calculates the pruning index to prune the sequence to with a given tolerance

    Parameters
    ----------
    data: pd.DataFrame
        Dataframe containing the pruning algorithm information

    tolerance: str
        Tolerance to prun the sequence

    Returns
    -------
    int
    """
    data = data[data['Coalition'] == 'Sum of contribution of events \u2264 t']
    if tolerance == 0:
        # to filter float unprecision out
        tolerance = 0.00000000001
    respecting_lens = data[data['Shapley Value'].abs() <= tolerance]
    if respecting_lens.shape[0] == 0:
        return -data['t (event index)'].min()

    return respecting_lens.iloc[0]['t (event index)']


def temp_coalition_pruning(f: Callable,
                           data: np.ndarray,
                           baseline: Union[np.ndarray, pd.DataFrame],
                           tolerance: float = None,
                           ret_plot_data=False,
                           verbose=False,
                           ) -> Union[int, pd.DataFrame, Tuple[int, pd.DataFrame]]:
    """Temporal coalition pruning method

    Parameters
    ----------
    f: Callable[[np.ndarray], np.ndarray]im
        Point of entry for model being explained.
        This method receives a 3-D np.ndarray (#samples, #seq_len, #features).
        This method returns a 2-D np.ndarray (#samples, 1).

    data: numpy.ndarray
        Input matrix to use. First element of the first dimension is explained,
        using the rest of the elements as context/hidden state.

    baseline: Union[np.ndarray, pd.DataFrame],
        Dataset baseline. Median/Mean of numerical features and mode of categorical.
        In case of np.array feature are assumed to be in order with `model_features`.
        The baseline can be an average event or an average sequence

    tolerance: float
        Temporal coalition explainer tolerance.
        Represents the maximum allowed Shapley Value of the older grouped events.

    ret_plot_data: bool
        If method returns pruning algorithm across the whole sequence

    verbose: bool
        If process is verbose

    Returns
    -------
    Union[int, pd.DataFrame, Tuple[int, pd.DataFrame]]:
        int:
            Pruning index
        pd.DataFrame
            Pruning data over the whole sequence
         Tuple[int, pd.DataFrame]]
            Pruning index and Pruning data over the whole sequence
    """
    if verbose:
        print("Allowed importance for pruned events: {}".format(tolerance))

    if ret_plot_data:
        plot_data = []
    pruning_idx = 0
    for seq_len in range(data.shape[1], -1, -1):
        explainer = TimeShapKernel(f, baseline, 0, "pruning")
        shap_values = explainer.shap_values(data, pruning_idx=seq_len, **{'nsamples': 4})
        if ret_plot_data:
            plot_data += [['Sum of contribution of events \u003E t', -data.shape[1]+seq_len, shap_values[0]]]
            plot_data += [['Sum of contribution of events \u2264 t', -data.shape[1]+seq_len, shap_values[1]]]

        if verbose:
            print("len {} | importance {}".format(-data.shape[1] + seq_len, shap_values[1]))

        if tolerance and seq_len == data.shape[1] and abs(shap_values[1]) <= tolerance:
            print("Unable to prune sequence.")

        if seq_len < data.shape[1] and tolerance and abs(shap_values[1]) <= tolerance:
            if pruning_idx == 0:
                pruning_idx = -data.shape[1] + seq_len
            if not ret_plot_data:
                return pruning_idx

    if tolerance is not None and pruning_idx == 0:
        pruning_idx = -data.shape[1]

    if tolerance is not None and ret_plot_data:
        # used for plotting
        return pruning_idx,  pd.DataFrame(plot_data, columns=['Coalition', 't (event index)', 'Shapley Value'])
    if tolerance is not None and not ret_plot_data:
        # used for event level
        return pruning_idx
    return pd.DataFrame(plot_data, columns=['Coalition', 't (event index)', 'Shapley Value'])


def local_pruning(f: Callable[[np.ndarray], np.ndarray],
                  data: np.ndarray,
                  pruning_dict: dict,
                  baseline: Union[np.ndarray, pd.DataFrame],
                  entity_uuid: Union[str, int, float] = None,
                  entity_col: str = None,
                  verbose: bool = False,
                  ) -> Tuple[pd.DataFrame, int]:
    """Method to prune a sequence or fetch the respective information if a path
    is provided

    Parameters
    ----------
    f: Callable[[np.ndarray], np.ndarray]
        Point of entry for model being explained.
        This method receives a 3-D np.ndarray (#samples, #seq_len, #features).
        This method returns a 2-D np.ndarray (#samples, 1).

    data: numpy.ndarray
        Input matrix to use. First element of the first dimension is explained,
        using the rest of the elements as context/hidden state.

    pruning_dict: dict
        Information required for pruning algorithm

    baseline: Union[np.ndarray, pd.DataFrame],
        Dataset baseline. Median/Mean of numerical features and mode of categorical.
        In case of np.array feature are assumed to be in order with `model_features`.
        The baseline can be an average event or an average sequence

    entity_uuid: Union[str, int, float]
        The indentifier of the sequence that is being pruned.
        Used when fetching information from a csv of explanations

    entity_col: str
        Column that contains the sequence identifiers
        Used when fetching information from a csv of explanations

    verbose: bool
        If process is verbose

    Returns
    -------
    Tuple[int, pd.DataFrame]]
            Pruning index and Pruning data over the whole sequence
    """
    def calculate_pruning():
        if baseline is None:
            raise ValueError("Baseline is not defined")
        coal_prun_idx, coal_plot_data = temp_coalition_pruning(f,
                                                               data,
                                                               baseline,
                                                               pruning_dict['tol'],
                                                               ret_plot_data=True,
                                                               verbose=verbose)

        return coal_prun_idx, coal_plot_data

    if pruning_dict.get("path") is None or not os.path.exists(pruning_dict.get("path")):
        #print("No path to explainer data provided. Calculating data")
        if baseline is None:
            raise ValueError("Baseline is not defined")
        coal_prun_idx, coal_plot_data = calculate_pruning()
        if pruning_dict.get("path") is not None:
            # create directory
            if '/' in pruning_dict.get("path"):
                Path(pruning_dict.get("path").rsplit("/", 1)[0]).mkdir(parents=True, exist_ok=True)
            coal_plot_data.to_csv(pruning_dict.get("path"), index=False)

    elif pruning_dict.get("path") is not None and os.path.exists(pruning_dict.get("path")):
        coal_plot_data = pd.read_csv(pruning_dict.get("path"))
        if len(coal_plot_data.columns) > 3:
            # global df
            assert entity_uuid is not None, "When using a dataset with several instances, a uuid needs to be provided"
            coal_plot_data = coal_plot_data[coal_plot_data[entity_col] == entity_uuid]
        coal_prun_idx = prune_given_data(coal_plot_data, pruning_dict.get('tol'))
    else:
        raise ValueError('Unrecognized explainer procedure.')
    return coal_plot_data, coal_prun_idx


def verify_pruning_dict(pruning_dict: dict):
    """Verifies the format of the pruning dict

    Parameters
    ----------
    pruning_dict: dict
    """
    if pruning_dict.get('path'):
        assert isinstance(pruning_dict.get('path'), str)

    assert pruning_dict.get('tol', False), "Tolerance(s) must be provided on the pruning dict"
    tolerances = pruning_dict.get('tol')
    if isinstance(tolerances, float):
        pruning_dict['tol'] = [tolerances]
    elif isinstance(tolerances, list):
        assert np.array([isinstance(x, float) for x in tolerances]).all(), "All provided tolerances must be floats."
    else:
        raise ValueError("Unsuported format of pruning tolerance(s). Please provide one tolerance or a list of them.")


def prune_all(f: Callable,
              data: Union[List[np.ndarray], pd.DataFrame, np.array],
              pruning_dict: dict,
              baseline: Union[pd.DataFrame, np.array] = None,
              model_features: List[Union[int, str]] = None,
              schema: List[str] = None,
              entity_col: Union[int, str] = None,
              time_col: Union[int, str] = None,
              append_to_files: bool = False,
              verbose: bool = False,
              ) -> pd.DataFrame:
    """Applies pruning to a dataset

    Parameters
    ----------
    f: Callable[[np.ndarray], np.ndarray]
        Point of entry for model being explained.
        This method receives a 3-D np.ndarray (#samples, #seq_len, #features).
        This method returns a 2-D np.ndarray (#samples, 1).

    data: Union[pd.DataFrame, np.array]
        Sequence to explain.

    pruning_dict: dict
        Information required for pruning algorithm

    baseline: Union[np.ndarray, pd.DataFrame],
        Dataset baseline. Median/Mean of numerical features and mode of categorical.
        In case of np.array feature are assumed to be in order with `model_features`.
        The baseline can be an average event or an average sequence

    model_features: List[str]
        In-order list of features to select and input to the model

    schema: List[str]
        Schema of provided data

    entity_col: str
        Column that contains the sequence identifiers

    time_col: str
        Data column that represents the time feature in order to sort sequences
        temporally

    append_to_files: bool
        Append explanations to files if file already exists

    verbose: bool
        If process is verbose

    Returns
    -------
    pd.DataFrame
    """
    if schema is None and isinstance(data, pd.DataFrame):
        schema = list(data.columns)
    verify_pruning_dict(pruning_dict)
    file_path = pruning_dict.get('path')
    tolerances = list(np.unique(pruning_dict.get('tol')))
    make_predictions = True
    prun_data = None
    if file_path is not None and os.path.exists(file_path) and not append_to_files:
        prun_data = pd.read_csv(file_path)
        make_predictions = False

        # TODO resume explanations
        # necessary_entities = set(np.unique(data[entity_col].values))
        # loaded_csv = pd.read_csv(file_path)
        # present_entities = set(np.unique(loaded_csv[entity_col].values))
        # if necessary_entities.issubset(present_entities):
        #     make_predictions = False
        #     prun_data = loaded_csv[loaded_csv[entity_col].isin(necessary_entities)]

    if make_predictions:
        ret_prun_data = []
        names = ["Coalition", "t (event index)", "Shapley Value", entity_col if isinstance(entity_col, str) else "Entity"]
        if file_path is not None:
            if os.path.exists(file_path):
                assert append_to_files, "The defined path for pruning data already exists and the append option is turned off. If you wish to append the explanations please use the flag `append_to_files`, otherwise change the provided path."
            else:
                if '/' in file_path:
                    Path(file_path.rsplit("/", 1)[0]).mkdir(parents=True, exist_ok=True)
                with open(file_path, 'w', newline='') as file:
                    writer = csv.writer(file, delimiter=',')
                    writer.writerow(names)

        if time_col is None:
            print("No time col provided, assuming dataset is ordered ascendingly by date")

        model_features_index, entity_col_index, time_col_index = convert_to_indexes(model_features, schema, entity_col, time_col)
        data = convert_data_to_3d(data, entity_col_index, time_col_index)

        for sequence in data:
            if entity_col is not None:
                entity = sequence[0, 0, entity_col_index]
            if model_features:
                sequence = sequence[:, :, model_features]
            sequence = sequence.astype(np.float64)
            local_pruning_data = temp_coalition_pruning(f, sequence, baseline, None, ret_plot_data=True, verbose=verbose)

            if entity_col is not None:
                local_pruning_data["Entity"] = entity

            ret_prun_data.append(local_pruning_data.values)
            if file_path is not None:
                 with open(file_path, 'a', newline='') as file:
                     writer = csv.writer(file, delimiter=',')
                     writer.writerows(local_pruning_data.values)

        prun_data = pd.DataFrame(np.concatenate(ret_prun_data), columns=names)

    df = calc_prun_indexes(prun_data, tolerances)
    return df



#  Copyright 2022 Feedzai
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import pandas as pd
from timeshap.plot import plot_temp_coalition_pruning, plot_event_heatmap, plot_feat_barplot, plot_cell_level
from timeshap.explainer import prune_given_data


def plot_local_report(pruning_dict: dict,
                      event_dict: dict,
                      feature_dict: dict,
                      cell_dict: dict,
                      coal_plot_data: pd.DataFrame = None,
                      event_data: pd.DataFrame = None,
                      feat_data: pd.DataFrame = None,
                      cell_data: pd.DataFrame = None,
                      ):
    """Plots a local report given explanations

    Parameters
    ----------
    pruning_dict: dict
        Information required for the pruning algorithm

    event_dict: dict
        Information required for the event level explanation calculation

    feature_dict: dict
        Information required for the feature level explanation calculation

    cell_dict: dict
        Information required for the cell level explanation calculation

    coal_plot_data: pd.DataFrame
        Pruning algorithm data to plot

    event_data: pd.DataFrame
        Event explanations to plot

    feat_data: pd.DataFrame
        Feature explanations to plot

    cell_data: pd.DataFrame
        Cell explanations to plot
    """
    if coal_plot_data is None:
        assert pruning_dict.get('path', False), "No data or path to data provided to calculate pruning statistics"
    if event_data is None:
        assert event_dict.get('path', False), "No data or path to data provided to plot event explanations"
    if feat_data is None:
        assert feature_dict.get('path', False), "No data or path to data provided to plot feature explanations"
    if cell_data is None and cell_dict is not None:
        assert cell_dict.get('path', False), "No data or path to data provided to plot feature explanations"

    if coal_plot_data is None:
        coal_plot_data = pd.read_csv(pruning_dict.get('path'))
    if event_data is None:
        event_data = pd.read_csv(event_dict.get('path'))
    if feat_data is None:
        feat_data = pd.read_csv(feature_dict.get('path'))
    if cell_data is None and cell_dict is not None:
        cell_data = pd.read_csv(cell_dict.get('path'))

    coal_prun_idx = prune_given_data(coal_plot_data, pruning_dict.get('tol'))
    plot_lim = max(abs(coal_prun_idx)+10, 40)
    pruning_plot = plot_temp_coalition_pruning(coal_plot_data, coal_prun_idx, plot_lim)

    event_plot = plot_event_heatmap(event_data)

    feature_plot = plot_feat_barplot(feat_data, feature_dict.get('top_feats'), feature_dict.get('plot_features'))

    if cell_dict:
        feat_names = list(feat_data['Feature'].values)[:-1]  # exclude pruned events
        cell_plot = plot_cell_level(cell_data, feat_names, feature_dict.get('plot_features'))
        plot_report = (pruning_plot | event_plot | feature_plot | cell_plot).resolve_scale(color='independent')
    else:
        plot_report = (pruning_plot | event_plot | feature_plot).resolve_scale(color='independent')

    return plot_report



#  Copyright 2022 Feedzai
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

# The MIT License (MIT)
#
# Copyright (c) 2018 Scott Lundberg
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
This file is based on the original SHAP implementation:
https://github.com/slundberg/shap/blob/master/shap/explainers/_kernel.py
"""

import numpy as np
import pandas as pd
import scipy as sp
import logging
import copy
import itertools
from typing import Tuple, Union

from timeshap.utils.timeshap_legacy import time_shap_match_instance_to_data, \
    time_shap_match_model_to_data, time_shap_convert_to_data, TimeShapDenseData
from shap.utils._legacy import convert_to_link, IdentityLink
from shap.utils._legacy import convert_to_instance, convert_to_model
from shap.explainers._kernel import Kernel
from scipy.special import binom
from scipy.sparse import issparse

log = logging.getLogger('shap')


class TimeShapKernel(Kernel):
    """Uses TimeSHAP Kernel SHAP method to explain the output of any function.

    TimeSHAP extends KernelSHAP to explain sequences of features on several axis.
    TimeSHAP calculates, event, feature, and cell level explanations.
    Due to sequences being arbitrarily long, TimeSHAP also implements a pruning
    algorithm based on Shapley values, to select the most relevant, recent,
    consecutive events.

    Parameters
    ----------
    model: function
        User supplied function that takes a 3D array (# samples x # sequence length x # features)
        and computes the output of the model for those samples. The output can be a vector
        (# samples) or a matrix (# samples x # model outputs).
        In order to use TimeSHAP in an optimized way, this model can also return the explained
        model's hidden state.

    background : numpy.array or pd.DataFrame
        The background event/sequence to use for integrating out features. To determine the impact
        of a feature, that feature is set to "missing" and the change in the model output
        is observed. Since most models aren't designed to handle arbitrary missing data at test
        time, we simulate "missing" by replacing the feature with the values it takes in the
        background dataset. So if the background dataset is a simple sample of all zeros, then
        we would approximate a feature being missing by setting it to zero.
        In TimeSHAP you can use an average event or average sequence.
        When using average events, consider using `timeshap.calc_avg_event` method to obtain it.
        When using average sequence, considering using `timeshap.calc_avg_sequence` method to obtain it.
        Note that when using the average sequence, all sequences of the dataset need to be the same.

    rs: int
        Random seed for timeshap algorithm

    mode: str
        This method indicates what kind of explanations should be calculated.
        Possible values: ["pruning", "event", "feature", "cell"]
            - "pruning" - used for pruning algorithm
            - "event" - used for event explanations
            - "feature" - used for feature explanations
            - "cell" -used for cell explanations

    varying: Tuple
        index of varying indexes on cell level
        If mode == "cell": varying needs to be of len 2, the first the idx of
            events to preturb, and the second the idx of features

    link : "identity" or "logit"
        A generalized linear model link to connect the feature importance values to the model
        output. Since the feature importance values, phi, sum up to the model output, it often makes
        sense to connect them to the output with a link function where link(output) = sum(phi).
        If the model output is a probability then the LogitLink link function makes the feature
        importance values have log-odds units.
    """
    def __init__(self, model, background, rs, mode, varying=None, link=IdentityLink(), **kwargs):
        self.background = background
        self.random_seed = rs
        self.mode = mode
        self.data = None
        self.varyingInds = None
        self.pruning_idx = None
        self.varying = varying
        self.returns_hs = None
        self.background_hs = None
        self.instance_hs = None
        if mode in 'cell':
            if varying is None:
                # The algorithm supports the calculation using all events and features
                # but its computation is very expensive as the number of cells is very large
                raise ValueError("Cell level needs to receive which cells to calculate")
            self.varying = varying
            cell_idx_keys = []
            for event_idx in self.varying[0]:
                for feat_idx in self.varying[1]:
                    cell_idx_keys += [[event_idx, feat_idx]]
            self.cell_idx_keys = np.array(cell_idx_keys)
            event_idx = self.cell_idx_keys[:, 0]
            self.considered_cells = {}
            for event in np.unique(event_idx):
                self.considered_cells[event] = self.cell_idx_keys[self.cell_idx_keys[:, 0] == event][:, 1]

        # convert incoming inputs to standardized iml objects
        self.link = convert_to_link(link)
        self.model = convert_to_model(model)
        self.keep_index = kwargs.get("keep_index", False)
        self.keep_index_ordered = kwargs.get("keep_index_ordered", False)

    def set_variables_up(self, X: np.ndarray):
        """Sets variables up for explanations

        Parameters
        ----------
        X: Union[pd.DataFrame, np.ndarray]
            Instance being explained
        """
        if len(self.background.shape) == 2:
            # 2D background needs to be expanded
            if self.background.shape[0] > 1 and not self.background.shape[0] == X.shape[1]:
                raise ValueError(
                    "When using background events, you can only pass one average event."
                    "When using background sequence, your background must be the same sequence length of the explained sequence")
            elif self.background.shape[0] > 1 and self.background.shape[0] == X.shape[1]:
                # average sequence
                sequence = copy.deepcopy(self.background)
            elif self.background.shape[0] == 1:
                # average event
                sequence = np.tile(self.background, (X.shape[1], 1))
            else:
                raise ValueError("Unknown combination of background and sequence sizes. Please open a ticket on github")
            sequence = np.expand_dims(sequence.copy(), axis=0)
        elif len(self.background.shape) == 3:
            if self.background.shape[1] > 1 and not self.background.shape[1] == X.shape[1]:
                raise ValueError(
                    "When using background events, you can only pass one average event."
                    "When using background sequence, your background must be the same sequence length of the explained sequence")
            elif self.background.shape[1] > 1 and self.background.shape[1] == X.shape[1]:
                # average sequence
                sequence = copy.deepcopy(self.background)
            elif self.background.shape[1] == 1:
                # average event
                sequence = np.tile(self.background, (1, X.shape[1], 1))
            else:
                raise ValueError("Unknown combination of background and sequence sizes. Please open a ticket on github")

        if self.mode == 'cell':
            self.data, self.special_cells = time_shap_convert_to_data(sequence, self.mode, self.pruning_idx, self.varying)
        else:
            self.data = time_shap_convert_to_data(sequence, self.mode, self.pruning_idx, self.varying)

        model_null, returns_hs = time_shap_match_model_to_data(self.model, self.data)
        self.returns_hs = returns_hs

        if not self.mode == 'pruning' and self.returns_hs:
            if self.pruning_idx == 0:
                # obtain the HS format
                _, example_hs = self.model.f(X[:, -1:, :])
                if isinstance(example_hs, tuple):
                    if isinstance(example_hs[0], tuple):
                        self.instance_hs = tuple(tuple(np.zeros_like(example_hs[y_i][i_x]) for i_x, x in enumerate(y)) for y_i, y in enumerate(example_hs))
                        self.background_hs = tuple(tuple(np.zeros_like(example_hs[y_i][i_x]) for i_x, x in enumerate(y)) for y_i, y in enumerate(example_hs))
                    else:
                        self.instance_hs = tuple(np.zeros_like(example_hs[i]) for i, x in enumerate(example_hs))
                        self.background_hs = tuple(np.zeros_like(example_hs[i]) for i, x in enumerate(example_hs))
                else:
                    self.instance_hs = np.zeros_like(example_hs)
                    self.background_hs = np.zeros_like(example_hs)
            else:
                _, self.background_hs = self.model.f(sequence[:, :self.pruning_idx, :])
                _, self.instance_hs = self.model.f(X[:, :self.pruning_idx, :])
                assert isinstance(self.background_hs, (np.ndarray, tuple)), "Hidden states are required to be numpy arrays or tuple "
                if isinstance(self.background_hs, tuple):
                    if isinstance(self.background_hs[0], tuple):
                        # working with LSTM
                        assert np.array([len(x) == 2 for x in self.background_hs]).all()
                        assert np.array([[isinstance(y, np.ndarray) for y in x] for x in self.background_hs]).all()
                    else:
                        assert np.array([isinstance(x, np.ndarray) for x in self.background_hs]).all()
        # enforce our current input type limitations
        assert isinstance(self.data, TimeShapDenseData), \
            "Shap explainer only supports the DenseData input currently."
        assert not self.data.transposed, "Shap explainer does not support transposed DenseData or SparseData currently."

        # warn users about large background data sets
        if len(self.data.weights) > 100:
            log.warning("Using " + str(len(
                self.data.weights)) + " background data samples could cause " +
                        "slower run times. Consider using shap.sample(data, K) or shap.kmeans(data, K) to " +
                        "summarize the background as K samples.")

        # init our parameters
        self.N = self.data.data.shape[0]
        # seq len total
        self.S = self.data.data.shape[1]
        self.P = self.data.data.shape[2]
        self.linkfv = np.vectorize(self.link.f)
        self.nsamplesAdded = 0
        self.nsamplesRun = 0

        # find E_x[f(x)]
        if isinstance(model_null, (pd.DataFrame, pd.Series)):
            model_null = np.squeeze(model_null.values)
        self.fnull = np.sum((model_null.T * self.data.weights).T, 0)
        self.expected_value = self.linkfv(self.fnull)

        # see if we have a vector output
        self.vector_out = True
        if len(self.fnull.shape) == 0:
            self.vector_out = False
            self.fnull = np.array([self.fnull])
            self.D = 1
            self.expected_value = float(self.expected_value)
        else:
            self.D = self.fnull.shape[0]

    def shap_values(self, X, pruning_idx=None, **kwargs):
        """ Estimate the SHAP values for a set of samples.

        Parameters
        ----------
        X : numpy.array
            A 3D matrix (#samples x #events x #features) on which to explain the model's output.

        nsamples : "auto" or int
            Number of times to re-evaluate the model when explaining each prediction. More samples
            lead to lower variance estimates of the SHAP values. The "auto" setting uses
            `nsamples = 2 * X.shape[1] + 2048`.

        l1_reg : "num_features(int)", "auto" (default for now, but deprecated), "aic", "bic", or float
            The l1 regularization to use for feature selection (the estimation procedure is based on
            a debiased lasso). The auto option currently uses "aic" when less that 20% of the possible sample
            space is enumerated, otherwise it uses no regularization. THE BEHAVIOR OF "auto" WILL CHANGE
            in a future version to be based on num_features instead of AIC.
            The "aic" and "bic" options use the AIC and BIC rules for regularization.
            Using "num_features(int)" selects a fix number of top features. Passing a float directly sets the
            "alpha" parameter of the sklearn.linear_model.Lasso model used for feature selection.

        Returns
        -------
        For models with a single output this returns a matrix of SHAP values
        (# samples x # features). Each row sums to the difference between the model output for that
        sample and the expected value of the model output (which is stored as expected_value
        attribute of the explainer). For models with vector outputs this returns a list
        of such matrices, one for each output.
        """
        assert isinstance(X, np.ndarray), "Instance must be 3D numpy array"
        if self.mode == "pruning":
            assert pruning_idx is not None
        else:
            assert pruning_idx < X.shape[1], "Pruning idx must be smaller than the sequence length. If not all events are pruned"
        assert pruning_idx % 1 == 0, "Pruning idx must be integer"
        self.pruning_idx = int(pruning_idx)

        self.set_variables_up(X)

        if sp.sparse.issparse(X) and not sp.sparse.isspmatrix_lil(X):
            X = X.tolil()

        # single instance
        if X.shape[0] == 1:
            explanation = self.explain(X, **kwargs)

            out = np.zeros(explanation.shape[0])
            if isinstance(explanation.shape, tuple) and len(explanation.shape) == 2:
                assert explanation.shape[1] == 1
                out[:] = explanation[:, 0]
            else:
                out[:] = explanation
            return out

    def explain(self, incoming_instance, **kwargs):
        # convert incoming input to a standardized iml object
        instance = convert_to_instance(incoming_instance)
        instance.group_display_values = self.data.group_names
        time_shap_match_instance_to_data(instance, self.data)

        # Find the feature groups we will test. If a feature does not change from its
        # current value then we know it doesn't impact the model
        if self.mode == "event":
            self.varyingInds = np.array([x for x in np.arange(incoming_instance.shape[1]-1, self.pruning_idx-1, -1)])
        elif self.mode == 'pruning':
            self.varyingInds = [0, 1]
        elif self.mode == "feature":
            if self.pruning_idx > 0:
                self.varyingInds = self.varying_groups(instance.x, self.data.groups_size - 1)
                # add an index for pruned events
                self.varyingInds = np.concatenate((self.varyingInds, np.array([self.data.groups_size - 1])))
            else:
                self.varyingInds = self.varying_groups(instance.x, self.data.groups_size)
        elif self.mode == 'cell':
            self.varyingInds = np.arange(len(self.data.groups))
        else:
            raise ValueError("`explain` -> mode not suported")

        if self.data.groups is None:
            self.varyingFeatureGroups = np.array([i for i in self.varyingInds])
            self.M = self.varyingFeatureGroups.shape[0]
        else:
            if self.mode in ['event']:
                self.varyingFeatureGroups = self.varyingInds
                self.M = len(self.varyingFeatureGroups)
                if self.pruning_idx > 0:
                    self.M += 1
            elif self.mode in ['feature']:
                if self.pruning_idx > 0:
                    self.varyingFeatureGroups = [self.data.groups[i] for i in self.varyingInds[:-1]]
                else:
                    self.varyingFeatureGroups = [self.data.groups[i] for i in self.varyingInds]
                self.M = len(self.varyingFeatureGroups)
                if self.pruning_idx > 0:
                    self.M += 1
            else:
                self.varyingFeatureGroups = [self.data.groups[i] for i in self.varyingInds]
                self.M = len(self.varyingFeatureGroups)

            groups = self.data.groups
            # convert to numpy array as it is much faster if not jagged array (all groups of same length)
            if isinstance(self.varyingFeatureGroups, list) and all(len(groups[i]) == len(groups[0]) for i in range(len(self.varyingFeatureGroups))):
                self.varyingFeatureGroups = np.array(self.varyingFeatureGroups)
                # further performance optimization in case each group has a single value
                if self.varyingFeatureGroups.shape[1] == 1:
                    self.varyingFeatureGroups = self.varyingFeatureGroups.flatten()

        if self.returns_hs:
            # Removed the input variability to receive pd.series and DataFrame
            model_out, _ = self.model.f(instance.x)
        else:
            model_out = self.model.f(instance.x)

        self.fx = model_out[0]
        if not self.vector_out:
            self.fx = np.array([self.fx])

        explained_score = (self.fx - self.fnull)[0]
        if abs(explained_score) < 0.1:
            raise ValueError(f"Score difference between baseline and instance ({explained_score}) is too low < 0.1."
                             f"Baseline score: {self.fx[0]} | Instance score: {self.fnull[0]}."
                             f"Consider choosing another baseline.")

        # if no features vary then no feature has an effect
        if self.M == 0:
            phi = np.zeros((self.data.groups_size, self.D))

        # if only one feature varies then it has all the effect
        elif self.M == 1:
            phi = np.zeros((self.data.groups_size, self.D))
            diff = self.link.f(self.fx) - self.link.f(self.fnull)
            for d in range(self.D):
                phi[self.varyingInds[0], d] = diff[d]

        # if more than one feature varies then we have to do real work
        else:
            self.l1_reg = kwargs.get("l1_reg", "auto")

            # pick a reasonable number of samples if the user didn't specify how many they wanted
            self.nsamples = kwargs.get("nsamples", "auto")
            if self.nsamples == "auto":
                self.nsamples = 2 * self.M + 2**11

            # if we have enough samples to enumerate all subsets then ignore the unneeded samples
            self.max_samples = 2 ** 30
            if self.M <= 30:
                self.max_samples = 2 ** self.M - 2
                if self.nsamples > self.max_samples:
                    self.nsamples = self.max_samples

            # reserve space for some of our computations
            self.allocate()

            # weight the different subset sizes
            num_subset_sizes = np.int(np.ceil((self.M - 1) / 2.0))
            num_paired_subset_sizes = np.int(np.floor((self.M - 1) / 2.0))
            weight_vector = np.array([(self.M - 1.0) / (i * (self.M - i)) for i in range(1, num_subset_sizes + 1)])
            weight_vector[:num_paired_subset_sizes] *= 2
            weight_vector /= np.sum(weight_vector)
            log.debug("weight_vector = {0}".format(weight_vector))
            log.debug("num_subset_sizes = {0}".format(num_subset_sizes))
            log.debug("num_paired_subset_sizes = {0}".format(num_paired_subset_sizes))
            log.debug("M = {0}".format(self.M))

            # fill out all the subset sizes we can completely enumerate
            # given nsamples*remaining_weight_vector[subset_size]
            num_full_subsets = 0
            num_samples_left = self.nsamples
            group_inds = np.arange(self.M, dtype='int64')
            mask = np.zeros(self.M)
            remaining_weight_vector = copy.copy(weight_vector)
            for subset_size in range(1, num_subset_sizes + 1):

                # determine how many subsets (and their complements) are of the current size
                nsubsets = binom(self.M, subset_size)
                if subset_size <= num_paired_subset_sizes: nsubsets *= 2
                log.debug("subset_size = {0}".format(subset_size))
                log.debug("nsubsets = {0}".format(nsubsets))
                log.debug("self.nsamples*weight_vector[subset_size-1] = {0}".format(
                    num_samples_left * remaining_weight_vector[subset_size - 1]))
                log.debug("self.nsamples*weight_vector[subset_size-1]/nsubsets = {0}".format(
                    num_samples_left * remaining_weight_vector[subset_size - 1] / nsubsets))

                # see if we have enough samples to enumerate all subsets of this size
                if num_samples_left * remaining_weight_vector[subset_size - 1] / nsubsets >= 1.0 - 1e-8:
                    num_full_subsets += 1
                    num_samples_left -= nsubsets

                    # rescale what's left of the remaining weight vector to sum to 1
                    if remaining_weight_vector[subset_size - 1] < 1.0:
                        remaining_weight_vector /= (1 - remaining_weight_vector[subset_size - 1])

                    # add all the samples of the current subset size
                    w = weight_vector[subset_size - 1] / binom(self.M, subset_size)
                    if subset_size <= num_paired_subset_sizes: w /= 2.0
                    for inds in itertools.combinations(group_inds, subset_size):
                        mask[:] = 0.0
                        mask[np.array(inds, dtype='int64')] = 1.0
                        self.add_sample(instance.x, mask, w)
                        if subset_size <= num_paired_subset_sizes:
                            mask[:] = np.abs(mask - 1)
                            self.add_sample(instance.x, mask, w)
                else:
                    break
            log.info("num_full_subsets = {0}".format(num_full_subsets))
            # add random samples from what is left of the subset space
            nfixed_samples = self.nsamplesAdded
            samples_left = self.nsamples - self.nsamplesAdded
            log.debug("samples_left = {0}".format(samples_left))
            np.random.seed(self.random_seed)
            if num_full_subsets != num_subset_sizes:
                remaining_weight_vector = copy.copy(weight_vector)
                remaining_weight_vector[:num_paired_subset_sizes] /= 2 # because we draw two samples each below
                remaining_weight_vector = remaining_weight_vector[num_full_subsets:]
                remaining_weight_vector /= np.sum(remaining_weight_vector)
                log.info("remaining_weight_vector = {0}".format(remaining_weight_vector))
                log.info("num_paired_subset_sizes = {0}".format(num_paired_subset_sizes))
                ind_set = np.random.choice(len(remaining_weight_vector), 4 * samples_left, p=remaining_weight_vector)
                ind_set_pos = 0
                used_masks = {}
                while samples_left > 0 and ind_set_pos < len(ind_set):
                    mask.fill(0.0)
                    ind = ind_set[ind_set_pos] # we call np.random.choice once to save time and then just read it here
                    ind_set_pos += 1
                    subset_size = ind + num_full_subsets + 1
                    mask[np.random.permutation(self.M)[:subset_size]] = 1.0

                    # only add the sample if we have not seen it before, otherwise just
                    # increment a previous sample's weight
                    mask_tuple = tuple(mask)
                    new_sample = False
                    if mask_tuple not in used_masks:
                        new_sample = True
                        used_masks[mask_tuple] = self.nsamplesAdded
                        samples_left -= 1
                        self.add_sample(instance.x, mask, 1.0)
                    else:
                        self.kernelWeights[used_masks[mask_tuple]] += 1.0

                    # add the compliment sample
                    if samples_left > 0 and subset_size <= num_paired_subset_sizes:
                        mask[:] = np.abs(mask - 1)

                        # only add the sample if we have not seen it before, otherwise just
                        # increment a previous sample's weight
                        if new_sample:
                            samples_left -= 1
                            self.add_sample(instance.x, mask, 1.0)
                        else:
                            # we know the compliment sample is the next one after the original sample, so + 1
                            self.kernelWeights[used_masks[mask_tuple] + 1] += 1.0

                # normalize the kernel weights for the random samples to equal the weight left after
                # the fixed enumerated samples have been already counted
                weight_left = np.sum(weight_vector[num_full_subsets:])
                log.info("weight_left = {0}".format(weight_left))
                self.kernelWeights[nfixed_samples:] *= weight_left / self.kernelWeights[nfixed_samples:].sum()

            # execute the model on the synthetic samples we have created
            self.run()

            # solve then expand the feature importance (Shapley value) vector to contain the non-varying features
            phi = np.zeros((self.data.groups_size, self.D))
            for d in range(self.D):
                vphi, _ = self.solve(self.nsamples / self.max_samples, d)
                if self.mode == 'event':
                    phi[:, d] = vphi
                elif self.mode == 'cell':
                    phi[:, d] = vphi
                else:
                    phi[self.varyingInds, d] = vphi

        if not self.vector_out:
            phi = np.squeeze(phi, axis=1)

        return phi

    @staticmethod
    def not_equal(i, j):
        if isinstance(i, str) or isinstance(j, str):
            return 0 if i == j else 1
        return 0 if np.isclose(i, j, equal_nan=True) else 1

    def varying_groups(self, x, group_size):
        if not sp.sparse.issparse(x):
            varying = np.zeros(group_size)
            for i in range(0, group_size):
                inds = self.data.groups[i]
                x_group = x[0, :, inds]
                if sp.sparse.issparse(x_group):
                    if all(j not in x.nonzero()[1] for j in inds):
                        varying[i] = False
                        continue
                    x_group = x_group.todense()
                num_mismatches = np.sum(np.frompyfunc(self.not_equal, 2, 1)(x_group, self.data.data[:, 0, inds]))
                varying[i] = num_mismatches > 0
            varying_indices = np.nonzero(varying)[0]
            return varying_indices
        else:
            # go over all nonzero columns in background and evaluation data
            # if both background and evaluation are zero, the column does not vary
            varying_indices = np.unique(np.union1d(self.data.data.nonzero()[1], x.nonzero()[1]))
            remove_unvarying_indices = []
            for i in range(0, len(varying_indices)):
                varying_index = varying_indices[i]
                # now verify the nonzero values do vary
                data_rows = self.data.data[:, [varying_index]]
                nonzero_rows = data_rows.nonzero()[0]

                if nonzero_rows.size > 0:
                    background_data_rows = data_rows[nonzero_rows]
                    if sp.sparse.issparse(background_data_rows):
                        background_data_rows = background_data_rows.toarray()
                    num_mismatches = np.sum(np.abs(background_data_rows - x[0, varying_index]) > 1e-7)
                    # Note: If feature column non-zero but some background zero, can't remove index
                    if num_mismatches == 0 and not \
                        (np.abs(x[0, [varying_index]][0, 0]) > 1e-7 and len(nonzero_rows) < data_rows.shape[0]):
                        remove_unvarying_indices.append(i)
            mask = np.ones(len(varying_indices), dtype=bool)
            mask[remove_unvarying_indices] = False
            varying_indices = varying_indices[mask]
            return varying_indices

    def allocate(self):
        if sp.sparse.issparse(self.data.data):
            # We tile the sparse matrix in csr format but convert it to lil
            # for performance when adding samples
            shape = self.data.data.shape
            nnz = self.data.data.nnz
            data_rows, data_cols = shape
            rows = data_rows * self.nsamples
            shape = rows, data_cols
            if nnz == 0:
                self.synth_data = sp.sparse.csr_matrix(shape, dtype=self.data.data.dtype).tolil()
            else:
                data = self.data.data.data
                indices = self.data.data.indices
                indptr = self.data.data.indptr
                last_indptr_idx = indptr[len(indptr) - 1]
                indptr_wo_last = indptr[:-1]
                new_indptrs = []
                for i in range(0, self.nsamples - 1):
                    new_indptrs.append(indptr_wo_last + (i * last_indptr_idx))
                new_indptrs.append(indptr + ((self.nsamples - 1) * last_indptr_idx))
                new_indptr = np.concatenate(new_indptrs)
                new_data = np.tile(data, self.nsamples)
                new_indices = np.tile(indices, self.nsamples)
                self.synth_data = sp.sparse.csr_matrix((new_data, new_indices, new_indptr), shape=shape).tolil()
        else:
            if self.returns_hs and self.mode != 'pruning':
                self.synth_data = np.tile(self.data.data[:, self.pruning_idx:, :], (self.nsamples, 1, 1))
                if isinstance(self.background_hs, tuple):
                    if isinstance(self.background_hs[0], tuple):
                        self.synth_hidden_states = tuple(tuple(np.tile(y, (1, self.nsamples, 1)) for y in x) for x in self.background_hs)
                    else:
                        self.synth_hidden_states = tuple(np.tile(x, (1, self.nsamples, 1)) for x in self.background_hs)
                else:
                    self.synth_hidden_states = np.tile(self.background_hs, (1, self.nsamples, 1))

            else:
                self.synth_data = np.tile(self.data.data, (self.nsamples, 1, 1))

        self.maskMatrix = np.zeros((self.nsamples, self.M))
        self.kernelWeights = np.zeros(self.nsamples)
        self.y = np.zeros((self.nsamples * self.N, self.D))
        self.ey = np.zeros((self.nsamples, self.D))
        self.lastMask = np.zeros(self.nsamples)
        self.nsamplesAdded = 0
        self.nsamplesRun = 0
        if self.keep_index:
            self.synth_data_index = np.tile(self.data.index_value, self.nsamples)

    def add_sample(self, x, m, w):
        offset = self.nsamplesAdded * self.N
        mask = m == 1.0
        if self.mode == "event":
            self.event_add_sample(x, mask, offset)
        elif self.mode == "feature":
            self.feat_add_sample(x, mask, offset)
        elif self.mode == 'pruning':
            self.pruning_add_sample(x, mask, offset)
        elif self.mode == "cell":
            self.cell_add_sample(x, mask, offset)
        else:
            raise ValueError("`add_sample` - Mode not supported")

        self.maskMatrix[self.nsamplesAdded, :] = m
        self.kernelWeights[self.nsamplesAdded] = w
        self.nsamplesAdded += 1

    def activate_background(self, x, offset):
        # in case self.pruning_idx == sequence length, we dont prune anything.
        if not self.pruning_idx == self.S:
            if self.returns_hs:
                # in case of using hidden state optimization, the background is the instance one
                if isinstance(self.synth_hidden_states, tuple):
                    if isinstance(self.synth_hidden_states[0], tuple):
                        for i, i_layer_state in enumerate(self.synth_hidden_states):
                            i_layer_state[0][:, offset:offset + self.N, :] = self.instance_hs[i][0]
                            i_layer_state[1][:, offset:offset + self.N, :] = self.instance_hs[i][1]
                    else:
                        for i, i_layer_state in enumerate(self.synth_hidden_states):
                            i_layer_state[:, offset:offset + self.N, :] = self.instance_hs[i]
                else:
                    self.synth_hidden_states[:, offset:offset + self.N, :] = self.instance_hs
            else:
                # in case of not using hidden state optimization, we need to set the whole background to the original sequence
                evaluation_data = x[0:1, :self.pruning_idx, :]
                self.synth_data[offset:offset + self.N, :self.pruning_idx, :] = evaluation_data

    def cell_add_sample(self, x, mask, offset):
        cells_to_preturb = self.cell_idx_keys[mask[: self.cell_idx_keys.shape[0]], :]
        relevent_events = np.unique(self.cell_idx_keys[:, 0])
        relevent_feats = np.unique(self.cell_idx_keys[:, 1])

        feats_by_event = {}
        for event in np.unique(cells_to_preturb[:, 0]):
            feats_by_event[event] = cells_to_preturb[cells_to_preturb[:, 0] == event][:, 1]

        # BACKGROUND IS ACTIVE
        if self.special_cells[3] and mask[-self.special_cells[3]]:
            self.activate_background(x, offset)

        # other cells are active (no relevant events or feats)
        if self.special_cells[2] and mask[-self.special_cells[2]]:
            other_events = [x for x in np.arange(self.pruning_idx, x.shape[1]) if x not in relevent_events]
            other_feats = [x for x in range(x.shape[2]) if x not in relevent_feats]

            for event in other_events:
                evaluation_data = x[0:1, event, other_feats]
                if self.returns_hs:
                    event = event - self.pruning_idx
                self.synth_data[offset:offset + self.N, event, other_feats] = evaluation_data

        mask_pointer = self.cell_idx_keys.shape[0]
        if self.special_cells[0]:
            # other feats in relevant events
            perturb_events = relevent_events[mask[mask_pointer: mask_pointer + len(self.varying[0])]]
            mask_pointer += len(self.varying[0])
            for event in perturb_events:
                other_feats = [x for x in range(x.shape[2]) if x not in relevent_feats]
                evaluation_data = x[0:1, event, other_feats]
                if self.returns_hs:
                    event = event - self.pruning_idx
                self.synth_data[offset:offset + self.N, event, other_feats] = evaluation_data

        if self.special_cells[1]:
            # other events in relevant feats
            perturb_feats = relevent_feats[mask[mask_pointer: mask_pointer + len(self.varying[1])]]
            mask_pointer += len(self.varying[1])
            other_events = [x for x in np.arange(self.pruning_idx, x.shape[1]) if x not in relevent_events]
            for event in other_events:
                evaluation_data = x[0:1, event, perturb_feats]
                if self.returns_hs:
                    event = event - self.pruning_idx
                self.synth_data[offset:offset + self.N, event, perturb_feats] = evaluation_data

        # activate individual cells
        for event, feats in feats_by_event.items():
            evaluation_data = x[0:1, event, feats]
            if self.returns_hs:
                event = event - self.pruning_idx
            self.synth_data[offset:offset + self.N, event, feats] = evaluation_data

    def event_add_sample(self, x, mask, offset):
        # there is a background and it is active
        if self.pruning_idx > 0 and mask[-1]:
            self.activate_background(x, offset)

        if self.pruning_idx > 0:
            # there is a background, so the last position of the mask is for it
            groups = self.varyingFeatureGroups[mask[:-1]]
        else:
            groups = self.varyingFeatureGroups[mask]

        evaluation_data = x[0:1, groups, :]
        if self.returns_hs:
            # re-align indexes to the truncated sequence
            groups = [x-self.pruning_idx for x in groups]
        self.synth_data[offset:offset + self.N, groups, :] = evaluation_data

    def feat_add_sample(self, x, mask, offset):
        #BACKGROUND IS ACTIVE
        if self.pruning_idx > 0 and mask[-1]:
            self.activate_background(x, offset)

        if self.pruning_idx > 0:
            # there is a background, so the last position of the mask is for it
            groups = self.varyingFeatureGroups[mask[:-1]]
        else:
            groups = self.varyingFeatureGroups[mask]

        evaluation_data = x[0:1, self.pruning_idx:, groups]
        if self.returns_hs:
            self.synth_data[offset:offset+self.N, :, groups] = evaluation_data
        else:
            self.synth_data[offset:offset+self.N, self.pruning_idx:, groups] = evaluation_data

    def pruning_add_sample(self, x, mask, offset):
        if not len(mask) == 2:
            raise ValueError("For pruning mode, masks must have size 2")
        if mask[0]:
            # cur active
            evaluation_data = x[0:1, self.pruning_idx:, :]
            self.synth_data[offset:offset + self.N, self.pruning_idx:, :] = evaluation_data
        if mask[1]:
            # background active
            evaluation_data = x[0:1, :self.pruning_idx, :]
            self.synth_data[offset:offset + self.N, :self.pruning_idx, :] = evaluation_data

    def run(self):
        num_to_run = self.nsamplesAdded * self.N - self.nsamplesRun * self.N

        data = self.synth_data[self.nsamplesRun * self.N:self.nsamplesAdded * self.N, :, :]

        if not self.mode == 'pruning' and self.returns_hs:
            if isinstance(self.synth_hidden_states, tuple):
                if isinstance(self.synth_hidden_states[0], tuple):
                    hidden_sates = tuple(tuple(y[:, self.nsamplesRun * self.N: self.nsamplesAdded * self.N,:] for y in x) for x in self.synth_hidden_states)
                else:
                    hidden_sates = tuple(x[:, self.nsamplesRun * self.N: self.nsamplesAdded * self.N,:] for x in self.synth_hidden_states)
            else:
                hidden_sates = self.synth_hidden_states[:, self.nsamplesRun * self.N: self.nsamplesAdded * self.N,:]

            modelOut, _ = self.model.f(data, hidden_sates)
        elif self.returns_hs:
            modelOut, _ = self.model.f(data)
        else:
            modelOut = self.model.f(data)

        if isinstance(modelOut, (pd.DataFrame, pd.Series)):
            modelOut = modelOut.values
        self.y[self.nsamplesRun * self.N:self.nsamplesAdded * self.N, :] = np.reshape(modelOut, (num_to_run, self.D))

        # find the expected value of each output
        for i in range(self.nsamplesRun, self.nsamplesAdded):
            eyVal = np.zeros(self.D)
            for j in range(0, self.N):
                eyVal += self.y[i * self.N + j, :] * self.data.weights[j]

            self.ey[i, :] = eyVal
            self.nsamplesRun += 1



#  Copyright 2022 Feedzai
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""Base/Parent Wrapper class.
"""

from abc import ABC


class TimeSHAPWrapper(ABC):
    """
    Base class for TimeSHAP model wrappers.
    """
    def __init__(self, model, batch_budget):
        self.model = model
        self.batch_budget = batch_budget

    def __call__(self, *args, **kwargs):
        return self.model(*args, *kwargs)



#  Copyright 2022 Feedzai
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import numpy as np
import pandas as pd
import altair as alt
import copy


def plot_temp_coalition_pruning(df: pd.DataFrame,
                                pruned_idx: int,
                                plot_limit: int = 50,
                                solve_negatives: bool = True,
                                ):
    """Plots the coalition explainer process

    Parameters
    ----------
    df: pd.DataFrame
        Pruning algorithm output

    pruned_idx: int
        Index the explainer takes place

    plot_limit: int
        Window of events to show the explainer to
        Default is 50

    solve_negatives: bool
        Whether to remove negative importances of the background instances
    """

    def solve_negatives_method(df):
        negative_values = copy.deepcopy(df[df['Shapley Value'] < 0])
        for idx, row in negative_values.iterrows():
            corresponding_row = df[np.logical_and(df['t (event index)'] == row['t (event index)'], ~(df['Coalition'] == row['Coalition']))]
            df.at[corresponding_row.index, 'Shapley Value'] = corresponding_row['Shapley Value'].values[0] + row['Shapley Value']
            df.at[idx, 'Shapley Value'] = 0
        return df

    df = df[df['t (event index)'] >= -plot_limit]
    if solve_negatives:
        df = solve_negatives_method(df)

    base = (alt.Chart(df).encode(
        x=alt.X("t (event index)", axis=alt.Axis(title='t (event index)', labelFontSize=15,
                              titleFontSize=15)),
        y=alt.Y("Shapley Value",
                axis=alt.Axis(titleFontSize=15, grid=True, labelFontSize=15,
                              titleX=-28),
                scale=alt.Scale(domain=[-0.05, 1], )),
        color=alt.Color('Coalition', scale=alt.Scale(
            domain=['Sum of contribution of events \u2264 t'],
            range=["#618FE0"]), legend=alt.Legend(title=None, labelLimit=0,
                                                  fillColor="white",
                                                  labelFontSize=14,
                                                  symbolStrokeWidth=0,
                                                  symbolSize=50,
                                                  orient="top-left")),
    )
    )

    area_chart = base.mark_area(opacity=0.5)
    line_chart = base.mark_line()
    line = alt.Chart(pd.DataFrame({'x': [pruned_idx]})).mark_rule(
        color='#E17560').encode(x='x')

    text1 = alt.Chart(pd.DataFrame({'x': [pruned_idx - 2]})).mark_text(
        text='Pruning', angle=270, color='#E17560', fontSize=15,
        fontWeight='bold').encode(x='x')

    pruning_graph = (area_chart + line_chart + line + text1).properties(width=350,height=225)

    return pruning_graph



#  Copyright 2022 Feedzai
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import pandas as pd
import numpy as np
import altair as alt
import re
import copy
from typing import List
import math


def plot_cell_level(cell_data: pd.DataFrame,
                    model_features: List[str],
                    plot_features: dict,
                    plot_parameters: dict = None,
                    ):
    """Plots local feature explanations

    Parameters
    ----------
    cell_data: pd.DataFrame
        Cell level explanations

    model_features: int
        The number of feature to display.

    plot_features: dict
        Dict containing mapping between model features and display features

    plot_parameters: dict
        Dict containing optinal plot parameters
            'height': height of the plot, default 225
            'width': width of the plot, default 200
            'axis_lims': plot Y domain, default [-0.5, 0.5]
            'FontSize': plot font size, default 15
    """
    c_range = ["#5f8fd6",
               "#99c3fb",
               "#f5f5f5",
               "#ffaa92",
               "#d16f5b",
               ]
    unique_events = [x for x in np.unique(cell_data['Event'].values) if x not in ['Other Events', 'Pruned Events']]
    sort_events = sorted(unique_events, key=lambda x:  re.findall(r'\d+', x)[0], reverse=True)
    unique_feats = [x for x in np.unique(cell_data['Feature'].values) if x not in ['Other Features', 'Pruned Events']]
    if plot_features:
        plot_features = copy.deepcopy(plot_features)
        sort_features = [plot_features[x] for x in model_features if x in unique_feats]
        if 'Other Features' in np.unique(cell_data['Feature'].values):
            plot_features['Other Features'] = 'Other Features'

        if 'Pruned Events' in np.unique(cell_data['Feature'].values):
            plot_features['Pruned Events'] = 'Pruned Events'

        cell_data['Feature'] = cell_data['Feature'].apply(lambda x: plot_features[x])
    else:
        sort_features = [x for x in model_features if x in unique_feats]

    cell_data['rounded'] = cell_data['Shapley Value'].apply(lambda x: round(x, 3))
    cell_data['rounded_str'] = cell_data['Shapley Value'].apply(lambda x: '0.000' if round(x, 3) == 0 else str(round(x, 3)))
    cell_data['rounded_str'] = cell_data['rounded_str'].apply(lambda x: f'{x}0' if len(x) == 4 else x)

    filtered_cell_data = cell_data[~np.logical_and(cell_data['Event'] == 'Pruned Events', cell_data['Feature'] == 'Pruned Events')]

    if plot_parameters is None:
        plot_parameters = {}
    height = plot_parameters.get('height', 225)
    width = plot_parameters.get('width', 200)
    axis_lims = plot_parameters.get('axis_lim', [-.5, .5])
    fontsize = plot_parameters.get('FontSize', 15)

    c = alt.Chart().encode(
        y=alt.Y('Feature', axis=alt.Axis(domain=False, labelFontSize=fontsize, title=None), sort=sort_features),
    )

    a = c.mark_rect().encode(
        x=alt.X('Event', axis=alt.Axis(titleFontSize=15), sort=sort_events),
        color=alt.Color('rounded', title=None,
                        legend=alt.Legend(gradientLength=height,
                                          gradientThickness=10, orient='right',
                                          labelFontSize=fontsize),
                        scale=alt.Scale(domain=axis_lims, range=c_range))
    )
    b = c.mark_text(align='right', baseline='middle', dx=18, fontSize=15,
                    color='#798184').encode(
            x=alt.X('Event', sort=sort_events,
                    axis=alt.Axis(orient="top", title='Shapley Value', domain=False,
                                  titleY=height + 20, titleX=172, labelAngle=30,
                                  labelFontSize=fontsize, )),
            text='rounded_str',
    )

    cell_plot = alt.layer(a, b, data=filtered_cell_data).properties(
        width=math.ceil(0.8*width),
        height=height
    )

    if 'Pruned Events' in np.unique(cell_data['Event'].values):
        # isolate the pruned contribution
        df_prun = cell_data[np.logical_and(cell_data['Event'] == 'Pruned Events',cell_data['Feature'] == 'Pruned Events')]
        assert df_prun.shape == (1, 5)
        prun_rounded_str = df_prun.iloc[0]['rounded_str']
        prun_rounded = df_prun.iloc[0]['rounded']
        df_prun = pd.DataFrame([[["Pruned", "Events"], "Other features", prun_rounded, prun_rounded_str], ],
                               columns=['Event', 'Feature', 'rounded', 'rounded_str'])

        c = alt.Chart().encode(y=alt.Y('Feature',
                                       axis=alt.Axis(labels=False, domain=False,
                                                     title=None)), )

        a = c.mark_rect().encode(
            x=alt.X('Event', axis=alt.Axis(titleFontSize=fontsize)),
            color=alt.Color('rounded', title=None, legend=None,
                            scale=alt.Scale(domain=axis_lims, range=c_range))
        )
        b = c.mark_text(align='right', dx=18, baseline='middle', fontSize=fontsize,
                        color='#798184').encode(
            x=alt.X('Event',
                    axis=alt.Axis(labelOffset=24, labelPadding=30, orient="top",
                                  title=None, domain=False, labelAngle=0,
                                  labelFontSize=fontsize, )),
            text='rounded_str',
        )

        cell_plot_prun = alt.layer(a, b, data=df_prun).properties(
            width=width / 3,
            height=height
        )

        cell_plot = alt.hconcat(cell_plot_prun, cell_plot).resolve_scale(color='independent')
    return cell_plot



#  Copyright 2022 Feedzai
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import pandas as pd
import numpy as np
from timeshap.utils import make_list, calculate_list_intersection
import altair as alt
from typing import Tuple, Callable


def multi_plot_wrapper(explanation_data: pd.DataFrame,
                       method: Callable,
                       parameters: tuple
                       ):
    """

    Parameters
    ----------
    explanation_data: pd.DataFrame
        Explanation Data

    method: Callable
        Plotter method

    parameters: tuple
        Plot parameters
    """
    data_nsamples = list(np.unique(explanation_data["NSamples"].values))
    data_rs = list(np.unique(explanation_data["Random Seed"].values))
    data_tol = list(np.unique(explanation_data["Tolerance"].values))

    multi_plot = True if len(data_nsamples) > 1 or len(data_rs) > 1 or len(data_tol) > 1 else False
    final_plot = alt.vconcat()

    for tolerance in data_tol:
        for rs in data_rs:
            for nsamples in data_nsamples:
                filtered_data = filter_dataset(explanation_data, tolerance, rs, nsamples)
                param_plot = method(filtered_data, {}) if parameters is None else method(filtered_data, *parameters)

                if multi_plot:
                    param_plot.properties(
                        title=f"Parameters: NSamples={nsamples} | Random Seed={rs} | Pruning Tol= {tolerance}"
                    )
                final_plot &= param_plot

    return final_plot


def filter_dataset(data: pd.DataFrame,
                   tol: float,
                   rs: int,
                   nsamples: int
                   ) -> pd.DataFrame:
    """Filters explanation dataset given parameters

    Parameters
    ----------
    data: pd.DataFrame
        Explanation Data

    tol: float

    rs: int

    nsamples: int

    Returns
    -------
    pd.DataFrame
        Filtered explanations
    """
    filtered_data = data[
        (data['Tolerance'] == tol) & (data['Random Seed'] == rs) & (
                    data['NSamples'] == nsamples)
        ]
    return filtered_data


def find_parameters_to_plot(event_dict: dict,
                            feature_dict: dict,
                            event_data: pd.DataFrame,
                            feat_data: pd.DataFrame
                            ) -> Tuple[list, list, list]:
    """Finds parameters to plot given explanation data and user configs

    Parameters
    ----------
    event_dict: dict
        Information required for the event level explanation calculation

    feature_dict: dict
        Information required for the feature level explanation calculation

    event_data: pd.DataFrame
        Event explanations to plot

    feat_data: pd.DataFrame
        Feature explanations to plot

    Returns
    -------
    list
        tolerances to plot

    list
        random seeds to plot

    list
        nsamples to plot
    """
    event_data_nsamples = list(np.unique(event_data["NSamples"].values))
    event_data_rs = list(np.unique(event_data["Random Seed"].values))
    event_data_tol = list(np.unique(event_data["Tolerance"].values))

    event_nsamples = make_list(event_dict.get('nsamples', event_data_nsamples))
    event_rs = make_list(event_dict.get('rs', event_data_rs))
    event_tol = make_list(event_dict.get('tol', event_data_tol))

    allowed_event_nsamples = calculate_list_intersection(event_data_nsamples, event_nsamples)
    allowed_event_rs = calculate_list_intersection(event_data_rs, event_rs)
    allowed_event_tol = calculate_list_intersection(event_data_tol, event_tol)

    feat_data_nsamples = list(np.unique(feat_data["NSamples"].values))
    feat_data_rs = list(np.unique(feat_data["Random Seed"].values))
    feat_data_tol = list(np.unique(feat_data["Tolerance"].values))

    feature_nsamples = make_list(feature_dict.get('nsamples', feat_data_nsamples))
    feature_rs = make_list(feature_dict.get('rs', feat_data_rs))
    feature_tol = make_list(feature_dict.get('tol', feat_data_tol))

    allowed_feat_nsamples = calculate_list_intersection(feat_data_nsamples, feature_nsamples)
    allowed_feat_rs = calculate_list_intersection(feat_data_rs, feature_rs)
    allowed_feat_tol = calculate_list_intersection(feat_data_tol, feature_tol)

    return calculate_list_intersection(allowed_event_tol, allowed_feat_tol), \
           calculate_list_intersection(allowed_event_rs, allowed_feat_rs), \
           calculate_list_intersection(allowed_event_nsamples, allowed_feat_nsamples)



#  Copyright 2022 Feedzai
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from typing import Callable, Union, Tuple, List
import numpy as np
import pandas as pd
from timeshap.explainer.kernel import TimeShapKernel
import os
from pathlib import Path


def cell_top_events(event_data: pd.DataFrame,
                    event_threshold: float = None,
                    top_x_events: int = None,
                    **kwargs,
                    ) -> Tuple[list, list]:
    """Calculates the indexes and names of events to participate in the cell level
    given the conditions

    Parameters
    ----------
    event_data: pd.DataFrame
        Event level explanations.

    event_threshold: float
        The threshold to consider an event relevant.

    top_x_events: float
        Number of events to include as relevant

    Returns
    -------
    Tuple[list, list]
        Tuple containing two lists.
        The first list contains the event indexes ordered ascendingly.
        The second list contains the correspondent event names.
    """
    # order explanations by absolute contribution
    ordered_exp = event_data.iloc[(-event_data['Shapley Value'].abs()).argsort()].reset_index()

    top_events_idx = []
    for _, row in ordered_exp.iterrows():
        if row['Feature'] == 'Pruned Events':
            # we want to skip previous events
            continue
        if event_threshold is not None and abs(row['Shapley Value']) < event_threshold:
            # we have reached the sopping point
            break

        top_events_idx += [[-row['index'] -1, row['Feature']]]

        if top_x_events is not None and len(top_events_idx) == top_x_events:
            # we have added enough events
            break

    df = pd.DataFrame(top_events_idx, columns=['idx', 'name']).sort_values('idx')
    return list(df['idx'].values), list(df['name'].values)


def cell_top_feats(feat_data: pd.DataFrame,
                   feat_threshold: float = None,
                   top_x_feats: int = None,
                   **kwargs,
                   ) -> Tuple[list, list]:
    """Calculates the indexes and names of events to participate in the cell level
    given the conditions

    Parameters
    ----------
    feat_data: pd.DataFrame
        Feature level explanations.

    feat_threshold: float
        The threshold to consider a feature relevant.

    top_x_feats: float
        Number of features to include as relevant

    Returns
    -------
    Tuple[list, list]
        Tuple containing two lists.
        The first list contains the feature indexes ordered ascendingly.
        The second list contains the correspondent feature names.
    """
    ordered_feats = feat_data.iloc[(-feat_data['Shapley Value'].abs()).argsort()].reset_index()

    top_feats_idx = []
    for _, row in ordered_feats.iterrows():
        if row['Feature'] == 'Pruned Events':
            # we want to skip previous events
            continue
        if feat_threshold is not None and abs(row['Shapley Value']) < feat_threshold:
            # we have reached the sopping point
            break
        top_feats_idx += [[row['index'], row['Feature']]]
        if top_x_feats is not None and len(top_feats_idx) == top_x_feats:
            # we have added enough events
            break

    df = pd.DataFrame(top_feats_idx, columns=['feat', 'name']).sort_values('feat')
    return list(df['feat'].values), list(df['name'].values)


def considered_cells(event_data: pd.DataFrame,
                     feat_data: pd.DataFrame,
                     **kwargs,
                     ) -> Tuple[Tuple[List, List], Tuple[List, List]]:
    """Calculates the indexes and names of events to participate in the cell level
    given the conditions

    Parameters
    ----------
    event_data: pd.DataFrame
        Event level explanations.

    feat_data: pd.DataFrame
        Feature level explanations.

    Returns
    -------
    Tuple[Tuple[List, List], Tuple[List, List]]
        Events and features to include in cell level computations
        and their respective name
    """
    top_events_idx, top_events_names = cell_top_events(event_data, **kwargs)
    top_feats_idx, top_feats_names = cell_top_feats(feat_data, **kwargs)

    return (top_events_idx, top_feats_idx), (top_events_names,top_feats_names)


def cell_level(f: Callable,
               data: np.ndarray,
               baseline: Union[pd.DataFrame, np.ndarray],
               event_data: pd.DataFrame,
               feat_data: pd.DataFrame,
               random_seed: int,
               nsamples: int,
               cell_dict: dict,
               pruned_idx: int,
               model_feats=None,
               ) -> pd.DataFrame:
    """Cell level given relevant events and features

    Parameters
    ----------
    f : Callable
        Prediction method of model being explained.
        Will be called with a 3-D input

    data: numpy.ndarray
        Input matrix to use. First element of the first dimension is explained,
        using the rest of the elements as context/hidden state.

    baseline: numpy.ndarray
        Baseline event to use. Median of numerical and mode for categorical.

    event_data: pd.DataFrame
        Event level explanations.

    feat_data: pd.DataFrame
        Feature level explanations.

    random_seed: int
        Used random seed for the sampling process.

    nsamples: int
        The number of coalitions for TimeSHAP to sample.

    cell_dict: dict
        Information required for the cell level explanation calculation

    pruned_idx: int
        Index to prune the sequence. All events up to this point are grouped\

    model_feats: List
        The list of feature names.
        If none is provided, "Feature 1" format is used

    Returns
    -------
    pd.DataFrame
    """
    kwargs = {}
    if cell_dict.get('threshold', False):
        # single threshold for everything
        kwargs['event_threshold'] = cell_dict.get('threshold')
        kwargs['feat_threshold'] = cell_dict.get('threshold')
    elif cell_dict.get('event_threshold', False) and cell_dict.get('feat_threshold', False):
        kwargs['event_threshold'] = cell_dict.get('event_threshold')
        kwargs['feat_threshold'] = cell_dict.get('feat_threshold')
    elif cell_dict.get('top_x', False):
        kwargs['top_x_events'] = cell_dict.get('top_x')
        kwargs['top_x_feats'] = cell_dict.get('top_x')
    elif cell_dict.get('top_x_events', False) and cell_dict.get('top_x_feats', False):
        kwargs['top_x_events'] = cell_dict.get('top_x_events')
        kwargs['top_x_feats'] = cell_dict.get('top_x_feats')
    else:
        raise ValueError("No threshold condition provided for cell level")

    varying_cells, names = considered_cells(event_data, feat_data, **kwargs)

    negative_indexes = np.array([False if x >= 0 else True for x in varying_cells[0]])
    if any(negative_indexes):
        if not all(negative_indexes):
            raise ValueError("All indexes must be positive or negative. Not both")
        varying_cells = ([data.shape[1] + x for x in varying_cells[0]], varying_cells[1])

    explainer = TimeShapKernel(f, baseline, random_seed, "cell", varying=varying_cells)
    explanation = explainer.shap_values(data, pruning_idx=pruned_idx, nsamples=nsamples)

    ret_df_data = []
    i = 0
    for event in names[0]:
        for feat in names[1]:
            row = [event, feat, explanation[i]]
            i += 1
            ret_df_data += [row]

    if explainer.special_cells[0]:
        for event in names[0]:
            row = [event, 'Other Features', explanation[i]]
            i += 1
            ret_df_data += [row]

    if explainer.special_cells[1]:
        for feat in names[1]:
            row = ['Other Events', feat, explanation[i]]
            i += 1
            ret_df_data += [row]

    if explainer.special_cells[2]:
        ret_df_data += [["Other Events", "Other Features", explanation[i]]]
        i += 1
    if explainer.special_cells[3]:
        ret_df_data += [["Pruned Events", "Pruned Events", explanation[i]]]
    return pd.DataFrame(ret_df_data, columns=['Event', 'Feature',
                                              'Shapley Value']).sort_values(
        'Shapley Value', ascending=False)


def local_cell_level(f: Callable[[np.ndarray], np.ndarray],
                     data: Union[pd.DataFrame, np.array],
                     cell_dict: dict,
                     event_data: pd.DataFrame,
                     feat_data: pd.DataFrame,
                     entity_uuid,
                     entity_col,
                     baseline,
                     pruned_idx,
                     ) -> pd.DataFrame:
    """Method to calculate event level explanations or load them if path is provided

    Parameters
    ----------
    f: Callable[[np.ndarray], np.ndarray]
        Point of entry for model being explained.
        This method receives a 3-D np.ndarray (#samples, #seq_len, #features).
        This method returns a 2-D np.ndarray (#samples, 1).

    data: pd.DataFrame
        Sequence to explain.

    cell_dict: dict
        Information required for the cell level explanation calculation

    event_data: pd.DataFrame
        Event level explanations.

    feat_data: pd.DataFrame
        Feature level explanations.

    entity_uuid: Union[str, int, float]
        The indentifier of the sequence that is being pruned.
        Used when fetching information from a csv of explanations

    entity_col: str
        Entity column to identify sequences

    baseline: Union[np.ndarray, pd.DataFrame],
        Dataset baseline. Median/Mean of numerical features and mode of categorical.
        In case of np.array feature are assumed to be in order with `model_features`.

    pruned_idx: int
        Index to prune the sequence. All events up to this point are grouped

    Returns
    -------
    pd.DataFrame
    """

    if cell_dict.get("path") is None or not os.path.exists(cell_dict.get("path")):
        #print("No path to cell data provided. Calculating data")
        cell_data = cell_level(f, data, baseline, event_data, feat_data, cell_dict.get("rs"), cell_dict.get("nsamples"), cell_dict, pruned_idx)
        if cell_dict.get("path") is not None:
            # create directory
            if '/' in cell_dict.get("path"):
                Path(cell_dict.get("path").rsplit("/", 1)[0]).mkdir(parents=True, exist_ok=True)
            cell_data.to_csv(cell_dict.get("path"), index=False)
    elif cell_dict.get("path") is not None and os.path.exists(cell_dict.get("path")):
        cell_data = pd.read_csv(cell_dict.get("path"))
        if len(cell_data.columns) == 5 and entity_col is not None:
            cell_data = cell_data[cell_data[entity_col] == entity_uuid]
        elif len(cell_data.columns) == 4:
            pass
        else:
            # TODO
            # the provided csv should be generated by timeshap, by either
            # explaining the whole dataset with TODO or just the instance in question
            raise ValueError
    else:
        raise ValueError
    return cell_data



#  Copyright 2022 Feedzai
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Callable, Union, List, Tuple
import numpy as np
import pandas as pd
from timeshap.explainer import prune_all, event_explain_all, feat_explain_all
from timeshap.explainer.pruning import verify_pruning_dict
from timeshap.explainer.event_level import verify_event_dict
from timeshap.explainer.feature_level import verify_feature_dict

from timeshap.plot import plot_global_report
import os
from timeshap.utils import convert_to_indexes, convert_data_to_3d, validate_input


def validate_global_input(f: Callable[[np.ndarray], np.ndarray],
                          data: Union[pd.DataFrame, np.array],
                          pruning_dict: dict,
                          event_dict: dict,
                          feature_dict: dict,
                          baseline: Union[pd.DataFrame, np.array] = None,
                          model_features: List[Union[int, str]] = None,
                          schema: List[str] = None,
                          entity_col: Union[int, str] = None,
                          time_col: Union[int, str] = None,
                          append_to_files: bool = False,
                          verbose: bool = False,
                          ):
    """ Validates the inputs for global reports

    Parameters
    ----------
    f: Callable[[np.ndarray], np.ndarray]
        Point of entry for model being explained.
        This method receives a 3-D np.ndarray (#samples, #seq_len, #features).
        This method returns a 2-D np.ndarray (#samples, 1).

    data: Union[List[np.ndarray], pd.DataFrame, np.array]
        Sequences to be explained.
        Must contain columns with names disclosed on `model_features`.

    pruning_dict: dict
        Information required for the pruning algorithm

    event_dict: dict
        Information required for the event level explanation calculation

    feature_dict: dict
        Information required for the feature level explanation calculation

    baseline: Union[np.ndarray, pd.DataFrame],
        Dataset baseline. Median/Mean of numerical features and mode of categorical.
        In case of np.array feature are assumed to be in order with `model_features`.
        The baseline can be an average event or an average sequence

    model_features: List[str]
        Features to be used by the model. Requires same order as training dataset

    schema: List[str]
        Schema of provided data

    entity_col: str
        Entity column to identify sequences

    time_col: str
        Data column that represents the time feature in order to sort sequences
        temporally

    append_to_files: bool
        Append explanations to files if file already exists

    verbose: bool
        If process is verbose

    """
    validate_input(f, data, baseline, model_features, schema, entity_col, time_col)

    if isinstance(data, np.ndarray):
        if len(data.shape) == 2:
            assert entity_col is not None, "Entity column must be provided when using 2D numpy arrays as data"

    verify_pruning_dict(pruning_dict)
    verify_event_dict(event_dict)
    verify_feature_dict(feature_dict)
    if pruning_dict.get("path"):
        if os.path.exists(pruning_dict.get("path")) and not append_to_files:
            print("The defined path for pruning data already exists and the append option is turned off. TimeSHAP will only read from this file and will not create new explanation data")
    else:
        print("No path to persist pruning data provided.")

    if event_dict.get("path"):
        if os.path.exists(event_dict.get("path")) and not append_to_files:
            print("The defined path for event explanations already exists and the append option is turned off. TimeSHAP will only read from this file and will not create new explanation data")
    else:
        print("No path to persist event explanations provided.")

    if feature_dict.get("path"):
        if os.path.exists(feature_dict.get("path")) and not append_to_files:
            print("The defined path for feature explanations already exists and the append option is turned off. TimeSHAP will only read from this file and will not create new explanation data")
    else:
        print("No path to persist feature explanations provided.")


def calc_global_explanations(f: Callable[[np.ndarray], np.ndarray],
                             data: Union[pd.DataFrame, np.array],
                             pruning_dict: dict,
                             event_dict: dict,
                             feature_dict: dict,
                             baseline: Union[pd.DataFrame, np.array] = None,
                             model_features: List[Union[int, str]] = None,
                             schema: List[str] = None,
                             entity_col: Union[int, str] = None,
                             time_col: Union[int, str] = None,
                             append_to_files: bool = False,
                             max_instances: int = 10000,
                             verbose: bool = False,
                             ) -> Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]:
    """ Calculates global report explanations

    Parameters
    ----------
    f: Callable[[np.ndarray], np.ndarray]
        Point of entry for model being explained.
        This method receives a 3-D np.ndarray (#samples, #seq_len, #features).
        This method returns a 2-D np.ndarray (#samples, 1).

    data: Union[List[np.ndarray], pd.DataFrame, np.array]
        Sequences to be explained.
        Must contain columns with names disclosed on `model_features`.

    pruning_dict: dict
        Information required for the pruning algorithm

    event_dict: dict
        Information required for the event level explanation calculation

    feature_dict: dict
        Information required for the feature level explanation calculation

    baseline: Union[np.ndarray, pd.DataFrame],
        Dataset baseline. Median/Mean of numerical features and mode of categorical.
        In case of np.array feature are assumed to be in order with `model_features`.
        The baseline can be an average event or an average sequence

    model_features: List[str]
        Features to be used by the model. Requires same order as training dataset

    schema: List[str]
        Schema of provided data

    entity_col: str
        Entity column to identify sequences

    time_col: str
        Data column that represents the time feature in order to sort sequences
        temporally

    append_to_files: bool
        Append explanations to files if file already exists

    max_instances: int
        Max number of instances to explain

    verbose: bool
        If process is verbose

    Returns
    -------
    pd.DataFrame
        Global pruning algorithm information

    pd.DataFrame
        Global event explanations

    pd.DataFrame
        Global feature explanations
    """
    if schema is None and isinstance(data, pd.DataFrame):
        schema = list(data.columns)

    validate_global_input(
        f, data, pruning_dict, event_dict, feature_dict, baseline, model_features,
        schema, entity_col, time_col, append_to_files, verbose)

    model_features_index, entity_col_index, time_col_index = convert_to_indexes(model_features, schema, entity_col, time_col)
    data = convert_data_to_3d(data, entity_col_index, time_col_index)
    if len(data) > max_instances:
        selected_sequences = np.random.choice(np.arange(len(data)), max_instances, False)
        data = [data[idx] for idx in selected_sequences]
    print("Calculating pruning algorithm")
    prun_indexes = prune_all(f, data, pruning_dict, baseline, model_features_index, schema, entity_col_index, time_col_index, append_to_files, verbose)

    print("Calculating event data")
    event_data = event_explain_all(f, data, event_dict, prun_indexes, baseline, model_features_index, schema, entity_col_index, time_col_index, append_to_files, verbose)

    print("Calculating feat data")
    feat_data = feat_explain_all(f, data, feature_dict, prun_indexes, baseline, model_features_index, schema, entity_col_index, time_col_index, append_to_files, verbose)

    return prun_indexes, event_data, feat_data


def global_report(f: Callable[[np.ndarray], np.ndarray],
                  data: Union[pd.DataFrame, np.array],
                  pruning_dict: dict,
                  event_dict: dict,
                  feature_dict: dict,
                  baseline: Union[pd.DataFrame, np.array] = None,
                  model_features: List[Union[int, str]] = None,
                  schema: List[str] = None,
                  entity_col: Union[int, str] = None,
                  time_col: Union[int, str] = None,
                  append_to_files: bool = False,
                  max_instances: int = 10000,
                  verbose: bool = False,
                  ):
    """ Calculates the global report and plots it.\

    Parameters
    ----------
    f: Callable[[np.ndarray], np.ndarray]
        Point of entry for model being explained.
        This method receives a 3-D np.ndarray (#samples, #seq_len, #features).
        This method returns a 2-D np.ndarray (#samples, 1).

    data: Union[List[np.ndarray], pd.DataFrame, np.array]
        Sequences to be explained.
        Must contain columns with names disclosed on `model_features`.

    pruning_dict: dict
        Information required for the pruning algorithm

    event_dict: dict
        Information required for the event level explanation calculation

    feature_dict: dict
        Information required for the feature level explanation calculation

    baseline: Union[np.ndarray, pd.DataFrame],
        Dataset baseline. Median/Mean of numerical features and mode of categorical.
        In case of np.array feature are assumed to be in order with `model_features`.
        The baseline can be an average event or an average sequence

    model_features: List[str]
        Features to be used by the model. Requires same order as training dataset

    schema: List[str]
        Schema of provided data

    entity_col: str
        Entity column to identify sequences

    time_col: str
        Data column that represents the time feature in order to sort sequences
        temporally

    append_to_files: bool
        Append explanations to files if file already exists

    max_instances: int
        Max instances to use for global plots and explanations.
        Used to limit explanation dump file sizes and allow for feasable
        plot time

    verbose: bool
        If process is verbose

    Returns
    -------
    pd.DataFrame

    altair.plot
    """
    prun_indexes, event_data, feat_data = \
        calc_global_explanations(f, data, pruning_dict, event_dict,
                                 feature_dict, baseline, model_features,
                                 schema, entity_col, time_col, append_to_files,
                                 max_instances, verbose
                                 )

    prun_stats, global_plot = \
        plot_global_report(pruning_dict, event_dict, feature_dict, prun_indexes,
                           event_data, feat_data
                           )

    return prun_stats, global_plot



#  Copyright 2022 Feedzai
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import pandas as pd
import numpy as np
import copy
import altair as alt
from timeshap.plot.utils import multi_plot_wrapper


def plot_feat_barplot(feat_data: pd.DataFrame,
                      top_x_feats: int = 15,
                      plot_features: dict = None
                      ):
    """Plots local feature explanations

    Parameters
    ----------
    feat_data: pd.DataFrame
        Feature explanations

    top_x_feats: int
        The number of feature to display.

    plot_features: dict
        Dict containing mapping between model features and display features
    """
    feat_data = copy.deepcopy(feat_data)
    if plot_features:
        plot_features['Pruned Events'] = 'Pruned Events'
        feat_data['Feature'] = feat_data['Feature'].apply(lambda x: plot_features[x])

    feat_data['sort_col'] = feat_data['Shapley Value'].apply(lambda x: abs(x))

    if top_x_feats is not None and feat_data.shape[0] > top_x_feats:
        sorted_df = feat_data.sort_values('sort_col', ascending=False)
        cutoff_contribution = abs(sorted_df.iloc[4]['Shapley Value'])
        feat_data = feat_data[np.logical_or(feat_data['Explanation'] >= cutoff_contribution, feat_data['Explanation'] <= -cutoff_contribution)]

    a = alt.Chart(feat_data).mark_bar(size=15, thickness=1).encode(
        y=alt.Y("Feature", axis=alt.Axis(title="Feature", labelFontSize=15,
                                         titleFontSize=15, titleX=-61),
                sort=alt.SortField(field='sort_col', order='descending')),
        x=alt.X('Shapley Value', axis=alt.Axis(grid=True, title="Shapley Value",
                                            labelFontSize=15, titleFontSize=15),
                scale=alt.Scale(domain=[-0.1, 0.4])),
    )

    line = alt.Chart(pd.DataFrame({'x': [0]})).mark_rule(
        color='#798184').encode(x='x')

    feature_plot = (a + line).properties(
        width=190,
        height=225
    )
    return feature_plot


def plot_global_feat(feat_data: pd.DataFrame,
                     top_x_feats: int = 12,
                     threshold: float = None,
                     plot_features: dict = None,
                     plot_parameters: dict = None,
                     **kwargs
                     ):
    """ Plots global feature plots

    Parameters
    ----------
    feat_data: pd.DataFrame
        Feature explanations to plot

    top_x_feats: int
        The number of feature to display.

    threshold: float
        The minimum absolute importance that a feature needs to have to be displayed

    plot_features: dict
        Dict containing mapping between model features and display features

    plot_parameters: dict
        Dict containing optional plot parameters
            'height': height of the plot, default 280
            'width': width of the plot, default 288
            'axis_lims': plot Y domain, default [-0.2, 0.6]
            'FontSize': plot font size, default 13
    """
    def plot(feat_data, top_x_feats, threshold, plot_features, plot_parameters):
        avg_df = feat_data.groupby('Feature').mean()['Shapley Value']
        if threshold is None and len(avg_df) >= top_x_feats:
            sorted_series = avg_df.abs().sort_values(ascending=False)
            threshold = sorted_series.iloc[top_x_feats-1]
        if threshold:
            avg_df = avg_df[np.logical_or(avg_df <= -threshold, avg_df >= threshold)]
        feat_data = feat_data[feat_data['Feature'].isin(avg_df.index)][['Shapley Value', 'Feature']]

        if threshold:
            avg_df = avg_df.append(pd.Series([0], index=['(...)']))
            #some prunning happened
            feat_data = feat_data.append({'Feature': '(...)', 'Shapley Value': -0.6, }, ignore_index=True)
        feat_data['type'] = 'Shapley Value'

        for index, value in avg_df.items():
            if index == '(...)':
                feat_data = feat_data.append(
                    {'Feature': index, 'Shapley Value': None, 'type': 'Mean'},
                    ignore_index=True)
            else:
                feat_data = feat_data.append(
                    {'Feature': index, 'Shapley Value': value, 'type': 'Mean'},
                    ignore_index=True)

        sort_features = list(avg_df.sort_values(ascending=False).index)
        if plot_features:
            plot_features = copy.deepcopy(plot_features)
            plot_features['Pruned Events'] = 'Pruned Events'
            plot_features['(...)'] = '(...)'
            feat_data['Feature'] = feat_data['Feature'].apply(lambda x: plot_features[x])
            sort_features = [plot_features[x] for x in sort_features]

        if plot_parameters is None:
            plot_parameters = {}
        height = plot_parameters.get('height', 280)
        width = plot_parameters.get('width', 288)
        axis_lims = plot_parameters.get('axis_lim', [-0.2, 0.6])
        fontsize = plot_parameters.get('FontSize', 13)

        global_feats_plot = alt.Chart(feat_data).mark_point(stroke='white',
                                                             strokeWidth=.6).encode(
            x=alt.X('Shapley Value', axis=alt.Axis(title='Shapley Value', grid=True),
                    scale=alt.Scale(domain=axis_lims)),
            y=alt.Y('Feature:O',
                    sort=sort_features,
                    axis=alt.Axis(labelFontSize=fontsize, titleX=-51)),
            color=alt.Color('type',
                            scale=alt.Scale(domain=['Shapley Value', 'Mean'],
                                            range=["#618FE0", '#d76d58']),
                            legend=alt.Legend(title=None, fillColor="white",
                                              symbolStrokeWidth=0, symbolSize=50,
                                              orient="bottom-right")),
            opacity=alt.condition(alt.datum.type == 'Mean', alt.value(1.0),
                                  alt.value(0.1)),
            size=alt.condition(alt.datum.type == 'Mean', alt.value(70),
                               alt.value(30)),
        ).properties(
            width=width,
            height=height
        )
        return global_feats_plot

    return multi_plot_wrapper(feat_data, plot, (top_x_feats, threshold, plot_features, plot_parameters))



#  Copyright 2022 Feedzai
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""Wrapper to explain custom PyTorch RNN models.
"""

import numpy as np
import pandas as pd
import copy
import math
from typing import Tuple

import torch

from timeshap.wrappers import TimeSHAPWrapper


class TorchModelWrapper(TimeSHAPWrapper):
    """Wrapper for pytorch machine learning models.

    Encompasses necessary logic to utilize torch models as lambda functions
    required for TimeSHAP explanations.

    This wrapper is responsible to create torch tensors, sending them to the
    required device, batching processes, and obtained predictions from tensors.

    Attributes
    ----------
    model: torch.nn.Module
        Torch model to wrap. This model is required to receive a torch.Tensor
        of sequences and returning the score for each instance of each sequence.

    batch_budget: int
        The number of instances to score at a time. Needed to not overload
        GPU memory.
        Default is 750K. Equates to a 7GB batch

    device: torch.device

    Methods
    -------
    predict_last(data: pd.DataFrame, metadata: Matrix) -> list
        Creates explanations for each instance in ``data``.
    """
    def __init__(self,
                 model: torch.nn.Module,
                 batch_budget: int = 750000,
                 device: torch.device = None,
                 ):
        super().__init__(model, batch_budget)
        if device:
            self.model = self.model.to(device)

    def prepare_input(self, input):
        sequence = copy.deepcopy(input)
        if isinstance(sequence, pd.DataFrame):
            sequence = np.expand_dims(sequence.values, axis=0)
        elif len(sequence.shape) == 2 and isinstance(sequence, np.ndarray):
            sequence = np.expand_dims(sequence, axis=0)

        if not (len(sequence.shape) == 3 and isinstance(sequence, np.ndarray)):
            raise ValueError("Input type not supported")

        return sequence

    def predict_last_hs(self,
                     sequences: np.ndarray,
                     hidden_states: np.ndarray = None,
                     ) -> Tuple[np.ndarray, Tuple[np.ndarray]]:
        sequences = self.prepare_input(sequences)
        device = next(self.model.parameters()).device

        sequence_len = sequences.shape[1]
        batch_size = math.floor(self.batch_budget / sequence_len)
        batch_size = max(1, batch_size)

        if sequences.shape[0] <= batch_size:
            with torch.no_grad():
                self.model.train(False)
                data_tensor = torch.from_numpy(sequences.copy()).float().to(device)
                if hidden_states is not None:
                    if isinstance(hidden_states, tuple):
                        if isinstance(hidden_states[0], tuple):
                            # for LSTM
                            hidden_states_tensor = tuple(tuple(torch.from_numpy(y).float().to(device)for y in x) for x in hidden_states)
                        else:
                            hidden_states_tensor = tuple(torch.from_numpy(x).float().to(device) for x in hidden_states)
                    else:
                        hidden_states_tensor = torch.from_numpy(hidden_states).float().to(device)
                    predictions = self.model(data_tensor, hidden_states_tensor)
                else:
                    predictions = self.model(data_tensor)
                self.model.train(True)
                if not isinstance(predictions, tuple):
                    return predictions.cpu().numpy()
                elif isinstance(predictions, tuple) and len(predictions) == 2:
                    predictions, hs = predictions
                    if isinstance(hs, tuple):
                        if isinstance(hs[0], tuple):
                            # for LSTM
                            return predictions.cpu().numpy(), tuple(tuple(y.cpu().numpy() for y in x) for x in hs)
                        else:
                            return predictions.cpu().numpy(), tuple(x.cpu().numpy() for x in hs)
                    else:
                        return predictions.cpu().numpy(), hs.cpu().numpy()
                else:
                    raise NotImplementedError(
                        "Only models that return predictions or predictions + hidden states are supported for now.")

        else:
            return_scores = []
            return_hs = []
            hs = None
            for i in range(0, sequences.shape[0], batch_size):
                batch = sequences[i:(i + batch_size), :, :]
                batch_tensor = torch.from_numpy(batch.copy())
                with torch.no_grad():
                    self.model.train(False)
                    batch_tensor = batch_tensor.float().to(device)
                    if hidden_states is not None:
                        if isinstance(hidden_states, tuple):
                            if isinstance(hidden_states[0], tuple):
                                # for LSTM
                                batch_hs_tensor = tuple(tuple(torch.from_numpy(y[:, i:(i + batch_size),:].copy()).float().to(device).float().to(device) for y in x) for x in hidden_states)
                            else:
                                batch_hs_tensor = tuple(torch.from_numpy(x[:, i:(i + batch_size), :].copy()).float().to(device).float().to(device) for x in hidden_states)
                        else:
                            batch_hs_tensor = torch.from_numpy(hidden_states[:, i:(i + batch_size), :].copy()).float().to(device)
                        predictions = self.model(batch_tensor, batch_hs_tensor)
                    else:
                        predictions = self.model(batch_tensor)
                    self.model.train(True)

                    if not isinstance(predictions, tuple):
                        predictions = predictions.cpu()
                    elif isinstance(predictions, tuple) and len(predictions) == 2:
                        predictions, hs = predictions
                        predictions = predictions.cpu()
                        if isinstance(hs, tuple):
                            if isinstance(hs[0], tuple):
                                if return_hs == []:
                                    return_hs = [[[] for _ in x] for x in hs]
                                for ith, ith_layer_hs in enumerate(hs):
                                    return_hs[ith][0].append(ith_layer_hs[0].cpu().numpy())
                                    return_hs[ith][1].append(ith_layer_hs[1].cpu().numpy())
                            else:
                                if return_hs == []:
                                    return_hs = [[] for _ in hs]
                                for ith, ith_layer_hs in enumerate(hs):
                                    return_hs[ith].append(ith_layer_hs.cpu().numpy())
                        else:
                            hs = hs.cpu().numpy()
                            return_hs.append(hs)
                    else:
                        raise NotImplementedError(
                            "Only models that return predictions or predictions + hidden states are supported for now.")

                return_scores.append(predictions.numpy())
            if hs is None:
                return np.concatenate(tuple(return_scores), axis=0)

            if isinstance(hs, tuple):
                if isinstance(hs[0], tuple):
                    return_hs = tuple(tuple(np.concatenate(tuple(y), axis=1) for y in x) for x in return_hs)
                else:
                    return_hs = tuple(np.concatenate(tuple(x), axis=1) for x in return_hs)
            else:
                return_hs = np.concatenate(tuple(return_hs), axis=1)
            return np.concatenate(tuple(return_scores), axis=0), return_hs



#  Copyright 2022 Feedzai
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import pandas as pd
import copy
import re
import altair as alt
from timeshap.plot.utils import multi_plot_wrapper


def plot_event_heatmap(event_data: pd.DataFrame,
                       ):
    """Plots local event explanations

    Parameters
    ----------
    event_data: pd.DataFrame
        Event global explanations
    """
    event_data = copy.deepcopy(event_data)
    # extract digit to order df by - this is redundant but gives security to the method
    event_data['idx'] = event_data['Feature'].apply(lambda x: event_data.shape[0] if x == 'Pruned Events' else int(re.findall(r'\d+', x)[0]) - 1)
    event_data = event_data.sort_values('idx')[['Shapley Value', 'Feature']]

    c_range = ["#5f8fd6",
               "#99c3fb",
               "#f5f5f5",
               "#ffaa92",
               "#d16f5b",
               ]

    event_data['row'] = event_data['Feature'].apply(lambda x: event_data.shape[0] if x == 'Pruned Events' else -eval(x.split(':')[0][6:]))
    event_data['column'] = event_data['Feature'].apply(lambda x: 1)
    event_data['rounded'] = event_data['Shapley Value'].apply(lambda x: round(x, 3))
    event_data['rounded_str'] = event_data['Shapley Value'].apply(lambda x: '0.000' if round(x, 3) == 0 else str(round(x, 3)))
    event_data['rounded_str'] = event_data['rounded_str'].apply(lambda x: f'{x}0' if len(x) == 4 else x)

    c = alt.Chart().encode(
        y=alt.Y('Feature:O',
                axis=alt.Axis(domain=False, labelFontSize=15, title='Event',
                              titleFontSize=15, titleX=-49),
                sort=list(event_data['Feature'].values), ),
    )

    a = c.mark_rect().encode(
        x=alt.X('column:O',
                axis=alt.Axis(title='Shapley Value', titleFontSize=15)),
        color=alt.Color('rounded', title=None,
                        legend=alt.Legend(gradientLength=225,
                                          gradientThickness=10, orient='right',
                                          labelFontSize=15),
                        scale=alt.Scale(domain=[-.5, .5], range=c_range))
    )
    b = c.mark_text(align='right', baseline='middle', dx=18, fontSize=15,
                    color='#798184').encode(
        x=alt.X('column:O',
                axis=alt.Axis(labels=False, title='Shapley Value', domain=False,
                              titleX=43)),
        text='rounded_str',
    )

    event_plot = alt.layer(a, b, data=event_data).properties(
        width=60,
        height=225
    )
    return event_plot


def plot_global_event(event_data: pd.DataFrame,
                      plot_parameters: dict = None,
                      ):
    """Plots global event explanations

    Parameters
    ----------
    event_data: pd.DataFrame
        Event global explanations

    plot_parameters: dict
        Dict containing optional plot parameters
            'height': height of the plot, default 150
            'width': width of the plot, default 360
            'axis_lims': plot Y domain, default [-0.3, 0.9]
            't_limit': number of events to plot, default -20
    """
    def plot(event_data: pd.DataFrame, plot_parameters: dict):
        event_data = copy.deepcopy(event_data)
        event_data = event_data[event_data['t (event index)'] < 1]
        event_data = event_data[['Shapley Value', 't (event index)']]

        event_data['type'] = 'Shapley Value'

        avg_df = event_data.groupby('t (event index)').mean()['Shapley Value']

        for index, value in avg_df.items():
            event_data = event_data.append({'t (event index)': index, 'Shapley Value': value, 'type': 'Mean'}, ignore_index=True)

        if plot_parameters is None:
            plot_parameters = {}

        height = plot_parameters.get('height', 150)
        width = plot_parameters.get('width', 360)
        axis_lims = plot_parameters.get('axis_lim', [-0.3, 0.9])
        t_limit = plot_parameters.get('axis_lim', -20)

        event_data = event_data[event_data['t (event index)'] >= t_limit]
        event_data = event_data[event_data['Shapley Value'] >= axis_lims[0]]
        event_data = event_data[event_data['Shapley Value'] <= axis_lims[1]]

        global_event = alt.Chart(event_data).mark_point(stroke='white',
                                                      strokeWidth=.6).encode(
            y=alt.Y('Shapley Value', axis=alt.Axis(grid=True, titleX=-23),
                    title="Shapley Value", scale=alt.Scale(domain=axis_lims, )),
            x=alt.X('t (event index):O', axis=alt.Axis(labelAngle=0)),
            color=alt.Color('type',
                            scale=alt.Scale(domain=['Shapley Value', 'Mean'],
                                            range=["#48caaa", '#d76d58']),
                            legend=alt.Legend(title=None, fillColor="white",
                                              symbolStrokeWidth=0, symbolSize=50,
                                              orient="top-left")),
            opacity=alt.condition(alt.datum.type == 'Mean', alt.value(1.0),
                                  alt.value(0.2)),
            size=alt.condition(alt.datum.type == 'Mean', alt.value(70),
                               alt.value(30)),
        ).properties(
            width=width,
            height=height,
        )

        return global_event

    return multi_plot_wrapper(event_data, plot, (plot_parameters))



#  Copyright 2022 Feedzai
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

# The MIT License (MIT)
#
# Copyright (c) 2018 Scott Lundberg
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import numpy as np
from shap.utils._legacy import Instance, Model, Data


class TimeShapDenseData(Data):
    def __init__(self, data, mode, group_names, *args):
        self.groups = args[0] if len(args) > 0 and args[0] is not None else [np.array([i]) for i in range(len(group_names))]

        if mode in ["event", "feature", "pruning", "cell"]:
            self.weights = args[2] if len(args) > 1 else np.ones(data.shape[0])
            self.weights /= np.sum(self.weights)

            self.transposed = False
            self.group_names = group_names
            self.data = data
            self.groups_size = len(self.groups)

        else:
            raise ValueError("TimeShapDenseData - mode not supported")


def time_shap_match_instance_to_data(instance, data):
    assert isinstance(instance, Instance), "instance must be of type Instance!"

    if isinstance(data, TimeShapDenseData):
        if instance.group_display_values is None:
            instance.group_display_values = [instance.x[0, :, group[0]] if len(group) == 1 else "" for group in data.groups]
        assert len(instance.group_display_values) == len(data.groups)
        instance.groups = data.groups
    else:
        raise NotImplementedError("Type of data not supported")


def time_shap_match_model_to_data(model, data):
    assert isinstance(model, Model), "model must be of type Model!"
    data = data.data
    returns_hs = False
    try:
        out_val = model.f(data)
        if len(out_val) == 2:
            # model returns the hidden state aswell.
            # We can use this hidden state to make the algorithm more efficent
            # as we reduce the computation of all pruned events to a single hidden state
            out_val, _ = out_val
            returns_hs = True
    except:
        print("Provided model function fails when applied to the provided data set.")
        raise

    if model.out_names is None:
        if len(out_val.shape) == 1:
            model.out_names = ["output value"]
        else:
            model.out_names = ["output value " + str(i) for i in range(out_val.shape[0])]

    return out_val, returns_hs


def time_shap_convert_to_data(val, mode, pruning_idx, varying=None):
    if type(val) == np.ndarray:
        if mode == 'event':
            event_names = ["Event: {}".format(i) for i in np.arange(val.shape[1], pruning_idx, -1)]
            if pruning_idx > 0:
                event_names += ["Pruned Events"]
            return TimeShapDenseData(val, mode, event_names)
        elif mode == 'feature':
            feature_names = ["Feat: {}".format(i) for i in np.arange(val.shape[2])]
            if pruning_idx > 0:
                feature_names += ["Pruned Events"]
            return TimeShapDenseData(val, mode, feature_names)
        elif mode == 'cell':
            group_names = []
            for event_idx in varying[0]:
                for feat_idx in varying[1]:
                    group_names += ["({}, {})".format(event_idx, feat_idx)]

            used_index = 0
            special_names = []
            # check if there are pruned cells
            if pruning_idx > 0:
                special_names += ["Pruned Cells"]
                used_index += 1
                pruned_events = used_index
            else:
                pruned_events = False

            # check if there are other cells
            if not(len(varying[0]) == val.shape[1] - pruning_idx or len(varying[1]) == val.shape[2]):
                special_names += ["Other Cells"]
                used_index += 1
                all_other = used_index
            else:
                all_other = False

            if len(varying[0]) < val.shape[1] - pruning_idx:
                special_names += reversed(["Other events on feature {}".format(x) for x in varying[1]])
                used_index += 1
                other_event_rel_feat = used_index
            else:
                other_event_rel_feat = False

            if len(varying[1]) < val.shape[2]:
                special_names += reversed(["Other feats on event {}".format(x) for x in varying[0]])
                used_index += 1
                other_feat_rel_event = used_index
            else:
                other_feat_rel_event = False

            group_names += reversed(special_names)

            return TimeShapDenseData(val, mode, group_names), [other_feat_rel_event, other_event_rel_feat, all_other, pruned_events]
        elif mode == 'pruning':
            return TimeShapDenseData(val, mode, ["x", "hidden"])
        else:
            raise ValueError("`convert_to_data` - mode not supported")

    elif isinstance(val, Data):
        return val
    else:
        assert False, "Unknown type passed as data object: " + str(type(val))



#  Copyright 2022 Feedzai
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import pandas as pd
from timeshap.explainer import pruning_statistics
from timeshap.plot import plot_global_feat, plot_global_event
import altair as alt
from timeshap.plot import find_parameters_to_plot, filter_dataset


def plot_global_report(pruning_dict: dict,
                       event_dict: dict,
                       feature_dict: dict,
                       prun_indexes: pd.DataFrame = None,
                       event_data: pd.DataFrame = None,
                       feat_data: pd.DataFrame = None
                       ):
    """Plots global epxlanation plots

    Parameters
    ----------
    pruning_dict: dict
        Information required for the pruning algorithm

    event_dict: dict
        Information required for the event level explanation calculation

    feature_dict: dict
        Information required for the feature level explanation calculation

    prun_indexes: pd.DataFrame
        Global Pruning information

    event_data: pd.DataFrame
        Global event explanations to plot

    feat_data: pd.DataFrame
        Global feature explanations to plot
    """

    if prun_indexes is None:
        assert pruning_dict.get('path', False), "No data or path to data provided to calculate pruning statistics"
    if event_data is None:
        assert event_dict.get('path', False), "No data or path to data provided to plot event explanations"
    if feat_data is None:
        assert feature_dict.get('path', False), "No data or path to data provided to plot feature explanations"

    if prun_indexes is None:
        prun_indexes = pd.read_csv(pruning_dict.get('path'))
    if event_data is None:
        event_data = pd.read_csv(event_dict.get('path'))
    if feat_data is None:
        feat_data = pd.read_csv(feature_dict.get('path'))

    print("Calculating pruning indexes")
    pruning_stats = pruning_statistics(prun_indexes, pruning_dict.get('tol'))

    plot_tols, plot_rs, plot_nsamples = find_parameters_to_plot(event_dict, feature_dict, event_data, feat_data)

    final_plot = alt.vconcat()
    for tolerance in plot_tols:
        for rs in plot_rs:
            for nsamples in plot_nsamples:

                plot_event_data = filter_dataset(event_data, tolerance, rs, nsamples)
                event_global_plot = plot_global_event(plot_event_data)

                plot_feat_data = filter_dataset(feat_data, tolerance, rs, nsamples)
                feat_global_plot = plot_global_feat(plot_feat_data, **feature_dict)

                horizontal_plot = alt.hconcat(event_global_plot, feat_global_plot, center=True)

                horizontal_plot.properties(
                    title=f"Parameters: NSamples={nsamples} | Random Seed={rs} | Pruning Tol= {tolerance}"
                )
                final_plot &= horizontal_plot

    return pruning_stats, final_plot



#  Copyright 2022 Feedzai
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Callable, Union, List, Tuple
import numpy as np
import pandas as pd
from timeshap.plot import plot_local_report
from timeshap.explainer import local_pruning, local_event, local_feat, local_cell_level
from timeshap.utils import validate_input


def validate_local_input(f: Callable[[np.ndarray], np.ndarray],
                         data: np.array,
                         pruning_dict: dict,
                         event_dict: dict,
                         feature_dict: dict,
                         cell_dict: dict = None,
                         baseline: Union[pd.DataFrame, np.ndarray]=None,
                         model_features: List[Union[int, str]] = None,
                         entity_col: Union[str, int] = None,
                         time_col: Union[str, int] = None,
                         entity_uuid: str = None,
                         ):
    """Verifies for local inputs if inputs are according

    Parameters
    ----------
    f: Callable[[np.ndarray], np.ndarray]
        Point of entry for model being explained.
        This method receives a 3-D np.ndarray (#samples, #seq_len, #features).
        This method returns a 2-D np.ndarray (#samples, 1).

    data: np.array
        Sequences to be explained.
        Must contain columns with names disclosed on `model_features`.

    pruning_dict: dict
        Information required for the pruning algorithm

    event_dict: dict
        Information required for the event level explanation calculation

    feature_dict: dict
        Information required for the feature level explanation calculation

    model_features: List[str]
        Features to be used by the model. Requires same order as training dataset

    entity_col: str
        Entity column to identify sequences

    time_col: str
        Data column that represents the time feature in order to sort sequences
        temporally
    """
    def check_dict(dict_to_check, key, types, message):
        if dict_to_check.get(key):
            assert isinstance(dict_to_check.get(key), types), message

    validate_input(f, data, baseline, model_features, None, entity_col, time_col)

    if isinstance(data, pd.DataFrame):
        data_cols = set(data.columns)
        if model_features:
            assert set(model_features).issubset(data_cols), "When providing model features, these should be on the given DataFrame"
        else:
            print("Assuming all features are model features")
            assert entity_col is None, "Entity col provided but no model features provided"
            assert time_col is None, "Time col provided but no model features provided"
        if entity_col is not None:
            assert entity_col in data_cols, "When providing entity feature, these should be on the given DataFrame"
            assert len(np.unique(data[entity_col].values)) == 1, "For local report, provided data must contain one instance only"
        if time_col is not None:
            assert time_col in data_cols, "When providing time feature, these should be on the given DataFrame"
    else:
        assert len(data.shape) == 3, "Provided data must be an numpy array with 3 dimensions"
        assert data.shape[0] == 1, "For local report, provided data must contain one instance only"

    assert baseline is None or isinstance(baseline, (pd.DataFrame, np.ndarray)), "Baseline must be a pd.DataFrame or np.ndarrays"

    assert pruning_dict.get("tol") is not None, "Prunning dict must have tolerance attribute"
    assert isinstance(pruning_dict.get("tol"), (int, float)), "Provided tolerance must be a int or float"
    if isinstance(pruning_dict.get("tol"), int):
        assert pruning_dict.get("tol") == 0, "Provided tolerance must be a float or 0"

    check_dict(event_dict, 'rs', int, "Provided random seed must be a int")
    check_dict(event_dict, 'nsamples', int, "Provided nsamples must be a int")
    check_dict(feature_dict, 'rs', int, "Provided random seed must be a int")
    check_dict(feature_dict, 'nsamples', int, "Provided nsamples must be a int")
    check_dict(feature_dict, 'top_feats', int, "Provided top_feats must be a int")
    check_dict(feature_dict, 'plot_features', dict, "Provided plot_features must be a dict, mapping model features, to plot features")

    if cell_dict is not None:
        check_dict(cell_dict, 'rs', int, "Provided random seed must be a int")
        check_dict(cell_dict, 'nsamples', int, "Provided nsamples must be a int")

        # assert we have
        if 'threshold' in cell_dict or 'top_x' in cell_dict:
            provided = 'threshold' if 'threshold' in cell_dict else 'top_x'
            assert 'feat_threshold' not in cell_dict, f"Provided both feat_threshold and {provided}. Please only provide one"
            assert 'top_x_feats' not in cell_dict, f"Provided both top_x_feats and {provided}. Please only provide one"
            assert 'event_threshold' not in cell_dict, f"Provided both event_threshold and {provided}. Please only provide one"
            assert 'top_x_events' not in cell_dict, f"Provided both top_x_events and {provided}. Please only provide one"
        else:
            if not('feat_threshold' in cell_dict or 'top_x_feats' in cell_dict):
                raise ValueError("No way to determine relevant features for cell level")

            if not ('event_threshold' in cell_dict or 'top_x_events' in cell_dict):
                raise ValueError( "No way to determine relevant events for cell level")


def calc_local_report(f: Callable[[np.ndarray], np.ndarray],
                      data: Union[pd.DataFrame, np.array],
                      pruning_dict: dict,
                      event_dict: dict,
                      feature_dict: dict,
                      cell_dict: dict = None,
                      baseline: Union[pd.DataFrame, np.ndarray] = None,
                      model_features: List[Union[int, str]] = None,
                      entity_col=None,
                      entity_uuid=None,
                      time_col=None,
                      verbose=False,
                      ) -> Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame, pd.DataFrame]:
    """Calculates local explanations

    Parameters
    ----------
    f: Callable[[np.ndarray], np.ndarray]
        Point of entry for model being explained.
        This method receives a 3-D np.ndarray (#samples, #seq_len, #features).
        This method returns a 2-D np.ndarray (#samples, 1).

    data: Union[pd.DataFrame, np.array]
        Sequence to be explained.

    pruning_dict: dict
        Information required for pruning algorithm

    event_dict: dict
        Information required for the event level explanation calculation

    feature_dict: dict
        Information required for the feature level explanation calculation

    cell_dict: dict
        Information required for the cell level explanation calculation

    entity_uuid: Union[str, int, float]
        The indentifier of the sequence that is being pruned.
        Used when fetching information from a csv of explanations

    entity_col: str
        Entity column to identify sequences

    time_col: str
        Data column that represents the time feature in order to sort sequences
        temporally

    model_features: List[str]
        Features to be used by the model. Requires same order as training dataset

    baseline: Union[pd.DataFrame, np.array]
        Dataset baseline. Median/Mean of numerical features and mode of categorical.
        In case of np.array feature are assumed to be in order with `model_features`.

    verbose: bool
        If process is verbose

    Returns
    -------
    pd.DataFrame
        Local pruning algorithm data

    pd.DataFrame
        Local event explanations

    pd.DataFrame
        Local feature explanations

    pd.DataFrame
        Local cell explanations
    """
    validate_local_input(f, data, pruning_dict, event_dict, feature_dict, cell_dict,
                         baseline, model_features, entity_col, time_col, entity_uuid)
    # deals with given date being a DataFrame
    if isinstance(data, pd.DataFrame):
        if time_col is not None:
            data[time_col] = data[[time_col]].apply(pd.to_numeric)
            data = data.sort_values(time_col)
        if model_features is not None:
            data = data[model_features]
        else:
            data = data.values
        data = np.expand_dims(data.to_numpy().copy(), axis=0).astype(float)

    coal_plot_data, coal_prun_idx = local_pruning(f, data, pruning_dict, baseline, entity_uuid, entity_col, verbose)
    pruning_idx = data.shape[1] + coal_prun_idx

    event_data = local_event(f, data, event_dict, entity_uuid, entity_col, baseline, pruning_idx)

    feature_data = local_feat(f, data, feature_dict, entity_uuid, entity_col, baseline, pruning_idx)

    if cell_dict:
        cell_data = local_cell_level(f, data, cell_dict, event_data, feature_data, entity_uuid, entity_col, baseline, pruning_idx)
    else:
        cell_data = None

    return coal_plot_data, event_data, feature_data, cell_data


def local_report(f: Callable[[np.ndarray], np.ndarray],
                 data: Union[pd.DataFrame, np.array],
                 pruning_dict: dict,
                 event_dict: dict,
                 feature_dict: dict,
                 cell_dict: dict = None,
                 baseline: Union[pd.DataFrame, np.array] = None,
                 model_features: List[Union[str, int]] = None,
                 entity_col: str = None,
                 entity_uuid: str = None,
                 time_col: str = None,
                 verbose=False,
                 ):
    """Calculates local report and plots it.

    Parameters
    ----------
    f: Callable[[np.ndarray], np.ndarray]
        Point of entry for model being explained.
        This method receives a 3-D np.ndarray (#samples, #seq_len, #features).
        This method returns a 2-D np.ndarray (#samples, 1).

    data: Union[pd.DataFrame, np.array]
        Sequence to be explained.

    pruning_dict: dict
        Information required for pruning algorithm

    event_dict: dict
        Information required for the event level explanation calculation

    feature_dict: dict
        Information required for the feature level explanation calculation

    cell_dict: dict
        Information required for the cell level explanation calculation

    entity_uuid: Union[str, int, float]
        The indentifier of the sequence that is being pruned.
        Used when fetching information from a csv of explanations

    entity_col: str
        Entity column to identify sequences

    time_col: str
        Data column that represents the time feature in order to sort sequences
        temporally

    model_features: List[str]
        Features to be used by the model. Requires same order as training dataset

    baseline: Union[pd.DataFrame, np.array]
        Dataset baseline. Median/Mean of numerical features and mode of categorical.
        In case of np.array feature are assumed to be in order with `model_features`.

    verbose: bool
        If process is verbose
    """
    pruning_data, event_data, feature_data, cell_level = \
        calc_local_report(f, data, pruning_dict, event_dict, feature_dict,
                          cell_dict, baseline, model_features, entity_col,
                          entity_uuid, time_col, verbose
                          )
    plot = plot_local_report(pruning_dict, event_dict, feature_dict, cell_dict,
                             pruning_data, event_data, feature_data, cell_level
                             )
    return plot



from pathlib import Path
import re
from setuptools import setup, find_packages


def stream_requirements(fd):
    """For a given requirements file descriptor, generate lines of
    distribution requirements, ignoring comments and chained requirement
    files.
    """
    for line in fd:
        cleaned = re.sub(r'#.*$', '', line).strip()
        if cleaned and not cleaned.startswith('-r'):
            yield cleaned

# ---------------------------------------------------------------------------- #
#                                   Requirements                               #
# ---------------------------------------------------------------------------- #

ROOT_PATH = Path(__file__).parent
README_PATH = ROOT_PATH / 'README.md'
REQUIREMENTS_PATH = ROOT_PATH / 'requirements' / 'main.txt'
REQUIREMENTS_TEST_PATH = ROOT_PATH / 'requirements' / 'test.txt'

with REQUIREMENTS_PATH.open() as requirements_file:
    requirements = list(stream_requirements(requirements_file))

with REQUIREMENTS_TEST_PATH.open() as test_requirements_file:
    requirements_test = requirements[:]
    requirements_test.extend(stream_requirements(test_requirements_file))

# ---------------------------------------------------------------------------- #
#                                   Version                                    #
# ---------------------------------------------------------------------------- #
SRC_PATH = ROOT_PATH / 'src' / 'timeshap'
VERSION_PATH = SRC_PATH / 'version.py'

with VERSION_PATH.open('rb') as version_file:
    exec(version_file.read())

# ---------------------------------------------------------------------------- #
#                                   SETUP                                      #
# ---------------------------------------------------------------------------- #
setup(
    name='timeshap',
    version=__version__,
    description="KernelSHAP adaptation for recurrent models.",
    keywords=['explainability', 'TimeShap'],
    
    long_description=(README_PATH).read_text(),
    long_description_content_type="text/markdown",
    
    author="Feedzai",
    url="https://github.com/feedzai/timeshap",

    package_dir={'': 'src'},
    packages=find_packages('src', exclude=['tests', 'tests.*']),
    package_data={
        '': ['*.yaml, *.yml'],
    },
    include_package_data=True,

    python_requires='>=3.6.*',

    install_requires=requirements,

    zip_safe=False,

    test_suite='tests',
    tests_require=requirements_test,
)
